/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/


#include "nest_collection_manager.h"
#include <string>


// **********************************
// Single NeST collection object
// **********************************
NestCollection::NestCollection( const string &name,
								const string &path,
								ClassAd			*ad )
{
	Clear( );
	Init( name, path, ad );
}

// Create a collection from a ClassAd
NestCollection::NestCollection( ClassAd *ad )
{
	buildCollectionFromAd( ad );
}

// Empty constructor
NestCollection::NestCollection( void )
{
	Clear( );
}

// Destructor
NestCollection::~NestCollection( )
{
}

// Initialize a Collection
NestReplyStatus NestCollection::Init( 
	const string &name,
	const string &path,
	ClassAd	*ad )
{
	if ( ( name.length() == 0 ) || ( path.length() == 0 ) ) {
		return NEST_BAD_PARAMETERS;
	}
	this->name = name;
	this->path = path;
	this->ad = ad;
	if ( ad ) {
		string key = COLLECTION_KEY;
		ad->InsertAttr( "Name", name );
		ad->InsertAttr( "Path", path );
		ad->InsertAttr( "Type", key );
	}
	return NEST_SUCCESS;
}

void
NestCollection::Clear( void )
{
	this->name = "";
	this->path = "";
	this->mtime = 0;
	this->version = "";
	this->comment = "";
	this->ad = NULL;
	this->valid = false;
}

// Build a collection from a ClassAd
bool
NestCollection::buildCollectionFromAd( 
	ClassAd			*ad )
{

    string	key;
	string	str;
	string	usename;

	// Reset the whole thing
	Clear( );

	// Extract the name from the ClassAd if none given
	key = "Name";
	if (!  ad->EvaluateAttrString( key, usename ) ) {
		return false;
	}

	// Extract the path
	key = "Path";
	if (!  ad->EvaluateAttrString( key, str ) ) {
		return false;
	}

	// Ok, we have a ligit collection; build it
	this->name = usename;
	this->path = str;
	this->ad = ad;
	this->valid = true;

	// Finally, extract the other fields & store 'em
	key = "Date";
	Value		val;
	if ( ad->EvaluateAttr( key, val ) ) {
		time_t	timeval;
		if ( val.IsIntegerValue( (int) timeval ) ) {
			setMtime( timeval );
		}
	}

	key = "Version";
	if ( ad->EvaluateAttrString( key, str ) ) {
		setVersion( str );
	}

	key = "Comment";
	if ( ad->EvaluateAttrString( key, str ) ) {
		setComment( str );
	}

	return true;
}

NestReplyStatus
NestCollection::setMtime( time_t mtime )
{
	this->mtime = mtime;
	if ( ad  && ( !ad->InsertAttr( "Date", (int) mtime ))  ) {
        return NEST_CLASSAD_ERROR;
	}
	return NEST_SUCCESS;
}

NestReplyStatus
NestCollection::setVersion( const string &version )
{
	this->version = version;
	if ( ad && ( ! ad->InsertAttr( "Version", version ))  ) {
        return NEST_CLASSAD_ERROR;
	}
	return NEST_SUCCESS;
}

NestReplyStatus
NestCollection::setVersion( const char *cversion )
{
	string	str;
	if ( cversion ) {
		str = (char *) cversion;
	}
	return setVersion( str );
}

NestReplyStatus
NestCollection::setComment( const string &comment )
{
	this->comment = comment;
	if ( ad && ( ! ad->InsertAttr( "comment", comment ))  ) {
        return NEST_CLASSAD_ERROR;
	}
	return NEST_SUCCESS;
}

NestReplyStatus
NestCollection::setComment( const char *ccomment )
{
	string	str;
	if ( ccomment ) {
		str = (char *)ccomment;
	}
	return setComment( str );
}


// **********************************
// Collection Manager
// **********************************

int
NestCollectionManager::init( )
{
	// not currently reading anything out of args
    return 0;
}

NestReplyStatus
NestCollectionManager::initialize( ClassAdCollection *resources )
{
	return setupView( resources );
}

NestReplyStatus
NestCollectionManager::recover( ClassAdCollection *resources )
{
	return setupView( resources );
}

// Setup the collection veiw
NestReplyStatus
NestCollectionManager::setupView( ClassAdCollection *resources )
{
	// Store ClassAdCollection information
	this->collection_list = resources;
	this->query.Bind( resources );
	this->collection_view = NEST_COLLECTION_VIEW;

	// Check to see if the "Collection" view already exists
	if ( ! this->collection_list->ViewExists( NEST_COLLECTION_VIEW ) ) {
		ViewName	parent = "root";
		string		view_rank("");
		string		view_partition_expr("");
		string 		view_constraint("");

		// Create the veiw here
		view_constraint = view_constraint + "other.Type == \"" + COLLECTION_KEY + "\"";
		if (!this->collection_list->CreateSubView( this->collection_view,
												   parent,
												   view_constraint,
												   view_rank,
												   view_partition_expr ) ) {
			nest_debug( NEST_DEBUG_ALWAYS,
						"NestCollectionManager Couldn't create view for itself!\n");
			return NEST_CLASSAD_ERROR;
		}
	}
	return NEST_SUCCESS;
}

NestReplyStatus
NestCollectionManager::createAssociation( const string &name,
                                          const string &path )
{
	// If it already exists, return an error
    if ( checkName( name ) ) {
        return NEST_ALREADY_EXISTS;
    }

	// Build a collection
	ClassAd			*ad = new ClassAd;
	NestCollection	*col = new NestCollection( name, path, ad );
	delete col;

	string	key = COLLECTION_KEY + name;
    if (!this->collection_list->AddClassAd( key, ad )) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "AddClassAd failed - key = %s - CondorErrMsg = %s \n",
                    key.c_str(),
                    CondorErrMsg.c_str() );

        return NEST_CLASSAD_ERROR;
    }

	
	
    return NEST_SUCCESS;
}

NestReplyStatus
NestCollectionManager::removeAssociation( const string &name )
{
    if ( ! checkName( name ) ) {
        return NEST_LOCAL_FILE_NOT_FOUND;
    }

    string key = COLLECTION_KEY + name;
    if (!this->collection_list->RemoveClassAd( key )) {
        return NEST_CLASSAD_ERROR;
	}
	return NEST_SUCCESS;
}

bool
NestCollectionManager::checkName( const string &name )
{
    string key = COLLECTION_KEY + name;

    ClassAd *ad = this->collection_list->GetClassAd( key );
    if (! ad) {
        return false;
    }

	return true;
}

NestCollection *
NestCollectionManager::getCollection( const string &name )
{
    string key = COLLECTION_KEY + name;

    ClassAd *ad = this->collection_list->GetClassAd(key);
    if (!ad) {
        return NULL;
    }

	// And, build a collection from it
    return new NestCollection( ad );
}

const string *
NestCollectionManager::getPath( const string &name )
{
    string key = COLLECTION_KEY;
    key = key + name;

    ClassAd *ad = this->collection_list->GetClassAd(key);
    if (!ad) {
        return NULL;
    }

	key = "Path";
	string	*str = new string;
	if (! ad->EvaluateAttrString( key, *str ) ) {
		return NULL;
	}

	return str;
}

NestReplyStatus
NestCollectionManager::updateMtime( NestCollection *col, time_t mtime )
{
	NestReplyStatus	status = col->setMtime( mtime );
	if ( status != NEST_SUCCESS ) {
		return status;
	}
	return update( col );
}

NestReplyStatus
NestCollectionManager::updateVersion( NestCollection *col, const string &version )
{
	NestReplyStatus	status = col->setVersion( version );
	if ( status != NEST_SUCCESS ) {
		return status;
	}
	return update( col );
}

NestReplyStatus
NestCollectionManager::updateComment( NestCollection *col, const string &comment )
{
	NestReplyStatus	status = col->setVersion( comment );
	if ( status != NEST_SUCCESS ) {
		return status;
	}
	return update( col );
}

NestReplyStatus
NestCollectionManager::update( NestCollection *col )
{
	ClassAd	*ad = col->getAd( );
	string tmp( (char *)col->getName( ) );
    string key = COLLECTION_KEY + tmp;
    if (! this->collection_list->UpdateClassAd( key, ad ) ) {
		return NEST_CLASSAD_ERROR;
	}
	return NEST_SUCCESS;
}

NestReplyStatus
NestCollectionManager::VisitAllCollections( NestCollectionVisitor &visitor )
{

	// Setup the query
    string constraint = "";
    ExprTree    *constraint_tree = this->parser.ParseExpression( constraint );

	// Setup the query
    if ( ! this->query.Query( this->collection_view, constraint_tree ) ) {
		nest_debug( NEST_DEBUG_ALWAYS, "Collection query failed: %s\n",
					CondorErrMsg.c_str() );
        return NEST_CLASSAD_ERROR;
	}
    this->query.ToFirst();
    string query_key;

	// And, start the query
    if (! this->query.Current( query_key ) ) {
		return NEST_SUCCESS;
    }

	// Loop through all of the matching ads
    do {
		ClassAd	*ad = this->collection_list->GetClassAd( query_key );
        if (! ad) {
			nest_debug( NEST_DEBUG_ALWAYS, "No ad\n" );
            continue;
        }
		NestCollection	*collection = new NestCollection( ad );
        if (! collection) {
			nest_debug( NEST_DEBUG_ALWAYS, "Failed building collection from Ad\n" );
            continue;
        }

		visitor.visit( *collection );
    } while ( this->query.Next( query_key ) );

	return NEST_SUCCESS;
}
