/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_concurrency_nob.h"

// constructor
nestConcurrencyNob::nestConcurrencyNob
        ( fd_set *readfds, fd_set *writefds, taskDoneCallback td_cb, int id ) 
    : nestConcurrencyModel( id ) // call base class constructor
{
    this->writefds = writefds;
    this->readfds  = readfds;
    this->maxfd    = 0;
    this->taskDone = td_cb;
}    

NestReplyStatus
nestConcurrencyNob::makeProgress(fd_set *r_readyfds, fd_set *w_readyfds) {

    TRANSFER_ITERATOR p; 
    nestNonblockingTransfer *transfer;
    bool removeTransfer;
    int status = 0;

    p = transfers.begin();
    // this can't be a for loop - because we may do removals 
    while( p != transfers.end() ) {
        transfer = *p;
        removeTransfer = false;
        int type = transfer->getType();
        
        if ( type == TEST_CLIENT_TO_SERVER || type == TEST_SERVER_TO_CLIENT ) {

                // handle test transfers differently
            if ( type == TEST_CLIENT_TO_SERVER ) {
                if ( NestFd_isset( transfer->getCsd(), r_readyfds ) ) {
                    transfer->readSrc();
                }
            } else {
                if ( NestFd_isset( transfer->getCsd(), w_readyfds ) ) {
                    transfer->writeDst();
                } 
            }
            status = transfer->getStatus();
            if ( status == WRITE_DONE || status == CLIENT_LOST 
                    || status == READ_DONE ) 
            {
                if ( type == TEST_CLIENT_TO_SERVER) {
                    NestFd_clear( transfer->getCsd(), readfds );
                } else {
                    NestFd_clear( transfer->getCsd(), writefds );
                }
                removeTransfer = true;
            }

        } else { 
                // foreach filetransfer, first send the ack to the client
                // then alternate between reading from src and writing to dst
            if ( ! transfer->ackSent() ) {
                if ( NestFd_isset( transfer->getCsd(), w_readyfds ) ) {
                    status = transfer->sendInitialAck();
                    NestFd_clear( transfer->getCsd(), writefds ); 
                    if ( status == CLIENT_LOST ) {
                        removeTransfer = true;
                    } else {
                        NestFd_set( transfer->getSrc(), readfds  );
                    }
                } else {
                    nest_debug(NEST_DEBUG_NOB, "Have to send an ACK but client not ready\n");
                }
    
            } else if ( NestFd_isset( transfer->getSrc(), r_readyfds ) ) {
                status = transfer->readSrc();
                NestFd_clear( transfer->getSrc(), readfds  );
                if ( status == CLIENT_LOST ) {
                    removeTransfer = true;
                } else {
                    NestFd_set( transfer->getDst(), writefds );
                }
            } else if ( NestFd_isset( transfer->getDst(), w_readyfds ) ) {
                status = transfer->writeDst();
                if ( status == WRITE_OK ) {
                    NestFd_clear( transfer->getDst(), writefds );
                    NestFd_set( transfer->getSrc(), readfds  );
                } else if ( status == WRITE_DONE || status == CLIENT_LOST ) {
                    NestFd_clear( transfer->getDst(), writefds );
                    removeTransfer = true;
                } else {
                    assert( status == WRITE_INCOMPLETE );
                }
            }
        }

        p++;

        if ( removeTransfer ) {
            NestReplyStatus nest_error = ( status != CLIENT_LOST ) 
                ? NEST_SUCCESS : NEST_NO_CONNECTION;
            endTransfer( transfer, nest_error );
        }
    }

    return NEST_SUCCESS;
}

NestReplyStatus
nestConcurrencyNob::endTransfer( nestNonblockingTransfer *transfer,
        NestReplyStatus nest_error ) 
{

    if ( transfer->getSrc() > 0 ) {
        assert( ! NestFd_isset( transfer->getSrc(), writefds ) );
        assert( ! NestFd_isset( transfer->getSrc(), readfds  ) );
    }
    if ( transfer->getDst() > 0 ) {
        assert( ! NestFd_isset( transfer->getDst(), writefds ) );
        assert( ! NestFd_isset( transfer->getDst(), readfds  ) );
    }

        // remove it from the list of currently managed transfers
    transfers.remove( transfer );
    if ( nest_error == NEST_SUCCESS ) {
        transfer->sendFinalAck( );
    } else {
        transfer->sendErrorMessage( __FILE__, __FUNCTION__,  nest_error );
    }
    transfer->endTransfer( (nest_error == NEST_SUCCESS) ); 

        // return responsibility to the main server
    taskDone( transfer->getClient(), nest_error, getId() );
    NestFd_set( transfer->getRealClientSocket(), readfds );
    delete( transfer );
    return NEST_SUCCESS;
}
    
char *
nestConcurrencyNob::toString() {
    static char name[] = "nob";
    return name;
}

int 
nestConcurrencyNob::getMaxfd( ) {
    return maxfd;
}

int 
nestConcurrencyNob::getConcurrencyCount( ) {
    return transfers.size();
}

NestReplyStatus
nestConcurrencyNob::addTask( nestClient* client ) {

    NestReplyStatus status = NEST_SUCCESS;

    if ( client->getConnection()->getType() == GFTP_PROTOCOL 
            || client->getConnection()->getType() == FTP_PROTOCOL ) {
        return NEST_NOT_YET_IMPLEMENTED;
    }


    nestNonblockingTransfer * noblock;
    noblock = nestNonblockingTransfer::newTransfer( client, status );
    if ( noblock == NULL ) return status;

    assert( ! NestFd_isset( noblock->getCsd(), writefds ) );
    assert( ! NestFd_isset( noblock->getCsd(), readfds ) );

    assert( client->getStorageConnection()->getSocket() );
    if ( client->getStorageConnection()->getSocket() > 0 ) {
            // a TEST will not have an fd
        assert(
           !NestFd_isset(client->getStorageConnection()->getSocket(),writefds));
        assert(
           !NestFd_isset(client->getStorageConnection()->getSocket(),readfds));
    }

    if ( noblock->getMaxfd() > maxfd ) maxfd = noblock->getMaxfd();
    transfers.push_back( noblock );        

        // set the inital filedescriptor for whatever is next
        // normal file transfers need to send an inital ack - tests do NOT
        // 3 pts also do not
        // PROXYS do not
    switch( noblock->getType() ) {
    case CLIENT_RECV_FILE:
    case CLIENT_SEND_FILE:
        NestFd_set( noblock->getCsd(), writefds );
        break;
    case THIRD_PARTY_PUT:
    case THIRD_PARTY_GET:
    case PROXY_GET:
    case PROXY_PUT:
        noblock->sendInitialAck();  /* this is a no-op here */
        NestFd_set( noblock->getSrc(), readfds );
        break;
    case TEST_CLIENT_TO_SERVER:
        noblock->sendInitialAck();  /* this is a no-op here */
        NestFd_set( noblock->getCsd(), readfds );
        break;
    case TEST_SERVER_TO_CLIENT:
        noblock->sendInitialAck();  /* this is a no-op here */
        NestFd_set( noblock->getCsd(), writefds );
        break;
    case COPY_REQUEST:
        NestFd_set( noblock->getSrc(), readfds );
        NestFd_set( noblock->getDst(), writefds );
        break;
    default:
        assert( 0 );
    }

    /*
        // fucking special case zero byte transfers
    if ( noblock->getLength() == 0 ) {
        NestFd_clear( noblock->getCsd(), writefds );
        NestFd_clear( noblock->getCsd(), readfds );
        NestFd_clear( noblock->getSrc(), readfds );
        status = endTransfer( noblock, NEST_SUCCESS ); 
    } 
    */

    return status;
}
