/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_concurrency_pot.h"

    // private helper functions
void * workerLoop( void * );
NestReplyStatus workerRequestHandler( int thread_id, nestClient *client );

bool doDirectCallback = false;

struct ThreadArgs {
    nestClientPool *pendingTasks;
    nestClientPool *doneTasks;
    int threadID;
    int poolSize;
    taskDoneCallback taskDone;
    pthread_mutex_t mutex;
    int server_pipe[2];
    int model_id;
};

    // constructor
nestConcurrencyPot::nestConcurrencyPot( 
        int pool_size, fd_set *readfds, taskDoneCallback cb, int id ) 
    : nestConcurrencyModel( id )   // call base class constructor
{

    thread_args = (struct ThreadArgs *)NestMalloc( sizeof(struct ThreadArgs) );
    thread_args->pendingTasks = new nestClientPool( true );
    thread_args->doneTasks    = new nestClientPool( true );

        // set up a pipe to communicate to server
    int pipe[2];
    if ( NestSpipe( pipe ) < 0 ) {
        nest_perror( "NestSpipe" );
        assert( 0 );
    }
    NestSetFileFlags( pipe[0], O_NDELAY );
    NestFd_set( pipe[0], readfds );
    maxfd = pipe[1];

    this->readfds = readfds;

        // set up the args
    thread_args->poolSize       = pool_size;
    thread_args->taskDone       = cb; 
    thread_args->server_pipe[0] = pipe[0];
    thread_args->server_pipe[1] = pipe[1];
    thread_args->model_id       = id;
    nest_pthread_mutex_init( &thread_args->mutex, NULL );

        // pass args to each newly created thread worker
    for ( thread_args->threadID = 0; thread_args->threadID < pool_size; 
            thread_args->threadID++ ) 
    {
        pthread_t *thread = (pthread_t *)NestMalloc( sizeof(*thread) );
        nest_pthread_create( thread, NULL, workerLoop, (void *)thread_args );
    }
}

nestConcurrencyPot::~nestConcurrencyPot( ) {
        // TODO:  make the workers go away somehow
    NestFree( thread_args );
}

bool
nestConcurrencyPot::directCallback() {
    return doDirectCallback;
}

char *
nestConcurrencyPot::toString() {
    static char name[] = "pot";
    return name;
}

int
nestConcurrencyPot::getMaxfd( ) {
    return maxfd;
}

NestReplyStatus
nestConcurrencyPot::addTask( nestClient *client ) {
    NestReplyStatus status = NEST_SUCCESS;
    thread_args->pendingTasks->addClient( client );
    return status;
}

NestReplyStatus
nestConcurrencyPot::makeProgress( fd_set *r_ready, fd_set *w_ready ) {
    NestReplyStatus status = NEST_SUCCESS;

        // check to see if the server has started ignoring us.
        // why would that happen?
    if ( ! NestFd_isset( thread_args->server_pipe[0], readfds ) ) {
        nest_die( "Somehow the dispatcher has decided to ignore "
                  "the thread pipe.\n" );
    }

        // clear the server signal pipe here so we can signal later
    if ( NestFd_isset( thread_args->server_pipe[0], r_ready ) ) {
        status = clearNestSignal();
        if ( status != NEST_SUCCESS ) {
            nest_warn( __FILE__, __FUNCTION__, "Unable to clear pot signal: %s\n", 
                    NestErrorString( status ) );
        } 
    }

    if ( ! doDirectCallback ) {
            // call taskDone only in admin thread
        while( thread_args->doneTasks->size() ) {
        //for( int i = 0; i < thread_args->doneTasks->size(); i++ ) {
            nestClient *client = thread_args->doneTasks->getFront();
            assert( client != NULL );
            NestReplyStatus status;
            taskFunction *fp = client->getTaskFunction();
            if ( fp != NULL ) {
                status = client->getTaskStatus();
            } else {
                assert( client->getRequest() != NULL );
                status = client->getRequest()->getStatus();
            }
            thread_args->taskDone( client, status, thread_args->model_id );
        }
    } else {
        assert( thread_args->doneTasks->size() == 0 );
    }
    return status;
}

// this function is called to make sure that server breaks out of its main
// select loop after a task has been completed
// otherwise the server will not be listening to the client
void
signalNest( int pipe, int id ) {
    static char msg[] = "POT wakeup\n";
    NestWriten( pipe, msg, NestStrlen( msg ) );    
    nest_debug( NEST_DEBUG_POT, "Thread %d signaled dispatcher.\n", id );
}

// this function is called after the server breaks out of the select loop
// just clears the signal so that it can be used again
NestReplyStatus
nestConcurrencyPot::clearNestSignal( ) {
    static char msg[FILEBUFFER];
    int n = 0;
    int clear = 0;
    nest_debug( NEST_DEBUG_POT, "Clearing nest signals from pot\n" );

    nest_pthread_mutex_lock( &thread_args->mutex );
    while (( n = NestRead( thread_args->server_pipe[0], msg, FILEBUFFER )) > 0){
        nest_debug( NEST_DEBUG_POT, "Cleared %d bytes: %s\n", n, msg );
        clear++;
    }
    nest_pthread_mutex_unlock( &thread_args->mutex );

        // if nothing was cleared return error
        // otherwise return success as long as errno is reasonable
    if ( clear == 0 ) return NEST_TEMPORARILY_UNAVAILABLE;
    return ( n == 0 || ( n == -1 && errno == EAGAIN ) )
            ? NEST_SUCCESS : ErrnoToNest( errno );
}

// this is the function where each worker thread lives
void *
workerLoop( void *void_args ) {

    struct ThreadArgs *args = ( struct ThreadArgs * ) void_args;
    NestReplyStatus status = NEST_SUCCESS;

    /*
    nest_debug( NEST_DEBUG_POT | NEST_DEBUG_ADMIN, 
            "Thread %d (pid %d) reporting\n", args->threadID, NestGetpid() );
    */
    
        // get a request from the queue, deal with it, repeat
    nestClient *currentClient;
    while( true ) {
        currentClient = args->pendingTasks->getFront( true );
        assert( currentClient != NULL );

        const char *request_type = ( currentClient->getRequest() == NULL )
                                 ? "special task"
                                 : currentClient->getRequest()->toString();

        nest_debug( NEST_DEBUG_POT, 
                "Thread %d handling request %s from client %s (%d)\n", 
                args->threadID, 
                request_type, 
                currentClient->getIdentity(),
                currentClient->getId() );
        status = workerRequestHandler( args->threadID, currentClient ); 
        nest_debug( NEST_DEBUG_POT, 
                "Thread %d handled request %s from client %s (%d): %s\n", 
                args->threadID, 
                request_type,
                currentClient->getIdentity(),
                currentClient->getId(), 
                NestErrorString( status ) );


            // return responsibility to the server and signal it, unlock mutex
        if ( doDirectCallback ) {
            nest_die( "I doubt doDirectCallback works - johnbent\n" );
                // lock the mutex and do shared cleanup and stats collection
            nest_pthread_mutex_lock( &args->mutex );
            args->taskDone( currentClient, status, args->model_id );
            nest_pthread_mutex_unlock( &args->mutex );
        } else {
            args->doneTasks->addClient( currentClient );
        }
        signalNest( args->server_pipe[1], args->threadID );
    }

}

NestReplyStatus
workerRequestHandler( int worker_id, nestClient *client ) {
    NestReplyStatus status = NEST_SUCCESS;

        // special case if client provides their own concurrency function
    taskFunction *fp = client->getTaskFunction();
    if ( fp != NULL ) {
        NestReplyStatus status = fp( client );
        nest_debug( NEST_DEBUG_PROXY, 
            "Thread %d handled special request from client %s(%d): %s\n", 
            worker_id, client->getIdentity(), client->getId(),
            NestErrorString( status ) );
        client->setTaskStatus( status );
        return status;
    }

    // TODO if 3pt, start it here
    // any 3pt is initiated here
    switch( client->getRequest()->getType() ) {
        case CLIENT_SEND_FILE:
                // TODO check error value
            client->sendInitialAck( );
            break;
        case CLIENT_RECV_FILE:
                // TODO check error value
            client->sendTransferInfo( );
            break;
        case THIRD_PARTY_GET:
        case THIRD_PARTY_PUT:
            nest_debug( NEST_DEBUG_POT, "Thread %d doing 3pt\n", worker_id);
            status =client->openThirdParty(client->getRequest()->getProtocol());
            break;
        case PROXY_GET:
        case PROXY_PUT:
                // nothing to do
            nest_debug( NEST_DEBUG_PROXY, "Thread %d starting proxy op.\n",
                    worker_id );
            break;
        case COPY_REQUEST:
            // nothing to do
            break; 
        default:
            assert( 0 );
            break;
    }

    if ( status != NEST_SUCCESS ) {
        client->getRequest()->setStatus( status );
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        return status;
    }

        // now go ahead and do the actual transfer 
    nest_debug( NEST_DEBUG_POT, "Thread %d enter blockingTask\n", worker_id );
    status = client->getTransfer()->BlockingTransfer( );
    nest_debug( NEST_DEBUG_POT, "Thread %d Task complete: %s\n", 
                worker_id, NestErrorString( status ) );

        // clean up after a 3pt
    if (   client->getRequest()->getType() == THIRD_PARTY_PUT
        || client->getRequest()->getType() == THIRD_PARTY_GET ) 
    {
        client->endThirdPartyTransfer( );
    }
        
    if ( status != NEST_SUCCESS ) {
        if ( status != NEST_LOT_NOT_ENOUGH ) {
            if ( errno != EPIPE ) {
                nest_warn( __FILE__, __FUNCTION__, "BlockingTask failure: %s.\n",
                        NestErrorString( status ) );
            }
            client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        }
    } else {
            // now end the transfer
        status = ( client->sendFinalAck() ? NEST_SUCCESS : NEST_NO_CONNECTION );
    }

    client->getRequest()->setStatus( status );
    return status;
}
