/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include <fstream>
#include "nest_common.h"
#include "nest_config.h"

enum 
HowSpecified {
    NotSpecified,
    SpecifiedEnvironment,
    SpecifiedCommandLine,
    SpecifiedConfigFile,
};

typedef struct config_variables
{
    NestConfigVariable  which_variable;
    char                *name;
    char                *short_arg;
    int                 defined;
    int                 requires_arg;
    char                *default_value;
    char                *usage;
    char                *help;
};

struct config_values
{
    bool defined;
    char *value;
    char *argument;
    char *short_arg;
    HowSpecified specified;
};

static struct config_values *config_values = NULL;
const char *prog_name = NULL;

static void StoreVariable(  config_variables *variable,
                            const char *value,
                            HowSpecified specified );

static void StoreVariable(const char *variable,
                          const char *value,
                          const char *config_file,
                          int line_number);

static bool ExtractVariable(string &line,
                            string &variable,
                            string &value,
                            int line_number);

#define CONFIG_VARIABLES_COUNT ( sizeof(config_variables) /\
                                 sizeof(struct config_variables))

#define NO_ARG_REQUIRED 0
#define ARG_REQUIRED 1
#define OPTIONAL_ARG 2

#define NO_ARG_DEFINED 0
#define ARG_DEFINED 1 

#define NO_ARGUMENT  0

struct config_variables config_variables[] =
{
    { Config_ConfigFile,
        "config_file",
        "c",
        NO_ARG_DEFINED,
        ARG_REQUIRED,
        "nest_config",
        "path",
        "Where NeST reads its config info",
    },
    { Config_MaintenanceInterval,             
        "maintenance_interval",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "60000",
        "#",
        "How frequently NeST should perform maintenance",
    },
    { Config_Debug,             
        "debug",
        "d",
        ARG_DEFINED,
        ARG_REQUIRED,
        "never",
        "value[,value]",
        "Turn on debug flags",
    },
    { Config_ProcessUser,             
        "user",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        ARG_REQUIRED,
        NULL,
        "user",
        "Which user id to run the NeST as",
    },
    { Config_Daemonize,        
        "daemon",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        NO_ARG_REQUIRED,
        NULL, 
        NULL,
        "Whether NeST should run as a daemon",
    },
    { Config_FakeCondorArg,        
        NO_ARGUMENT, 
        "f",
        NO_ARG_DEFINED,
        NO_ARG_REQUIRED,
        NULL,
        NULL,
        "This argument does nothing, used for compatibility with Condor master",
    },
    { Config_DisableAuthentication,        
        "noauth",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        NO_ARG_REQUIRED,
        NULL,
        NULL,
        "Whether NeST should run in totally open mode (DISCOURAGED)",
    },
    { Config_ThreadPoolSize,        
        "threadpool",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "25",
        "#",
        "How many worker threads should be used",
    },
    { Config_AdmissionControl,        
        "admission",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        ARG_REQUIRED,
        NULL,
        "#",
        "How many concurrent requests should NeST handle before queuing",
    },
    { Config_Bandwidth,        
        "bandwidth",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        ARG_REQUIRED,
        NULL,
        "#",
        "How much bandwidth (in MB) should NeST limit itself to",
    },
    { Config_EnableThreadModel,        
        "thread_model",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "1",
        "0|1",
        "Whether NeST should use threads for concurrency",
    },
    { Config_EnableProcessModel,        
        "process_model",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "0",
        "0|1",
        "Whether NeST should use processes for concurrency",
    },
    { Config_EnableEventModel,        
        "event_model",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "0",
        "0|1",
        "Whether NeST should use events for concurrency",
    },
    { Config_ConcurrencySelectionPolicy,        
        "concurrency_selection",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "lottery",
        "lottery|take-turns",
        "How should NeST select between different concurrency models",
    },
    { Config_CacheModelPolicy,        
        "cache_model",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "clock",
        "clock|lru",
        "Which replacement policy should NeST simulate on its cache model",
    },
    { Config_SchedulingPolicy,        
        "schedule",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "cache",
        "cache|fifo|sff",
        "When reqs get queued, how should NeST select the next to service",
    },
    { Config_PreferredProtocol,        
        "preferred_protocol",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        ARG_REQUIRED,
        NULL,
        "nfs|http|chirp|gftp",
        "When reqs get queued, should NeST prefer reqs from any protocol",
    },
    { Config_StorageDir,        
        "storage_dir",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "__NEST__",
        "path",
        "Where NeST should store it's files",
    },
	{ Config_Latency,
        "latency",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        ARG_REQUIRED,
        NULL,
        "#",
        "How much latency NeST should impose before servicing requests"
        " (used for testing to make a NeST appear remote)",
    },
	{ Config_ProxyHost,			
        "home_node",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        ARG_REQUIRED,
        NULL, 
        "host_name",
        "Which other NeST should this one act as a proxy for",
    },
    { Config_ProxysStandAlone,        
        "proxys_standalone",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        NO_ARG_REQUIRED,
        NULL, 
        NULL,
        "Whether a NeST proxy should NOT cooperate with other proxies",
    },
    { Config_DisableLots,       // hmmm.  how do I do this?
        "disable_lots",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        NO_ARG_REQUIRED,
        NULL,
        NULL,
        "Force NeST to run without lot management (discouraged)",
    },
    { Config_MaxLotDuration,    
        "max_lot_duration",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "525600000",      // 1000 years in minutes
        "#",
        "Maximum lot time in minutes",
    },
    { Config_DefaultLotDuration,
        "default_lot_duration",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "1440",          // one day in minutes
        "#",
        "Default lot time in minutes",
    },
    { Config_StorageSpace,      
        "storage_space",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        ARG_REQUIRED,
        NULL,
        "#",
        "Limit NeST to only use a fixed amount of disk (in MB)",
    },
    { Config_UserList,          
        "user_list",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        ARG_REQUIRED,
        NULL,
        "user[,user]",
        "Which users are allowed to connect to the NeST"
            " (if unspecified, uses grid mapfile)",
    },
    { Config_DenyList,
        "deny_list",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        ARG_REQUIRED,
        NULL,
        "user[,user]",
        "Which users are denied access to the NeST",
    },
    { Config_AdminUser,         
        "admin_user",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "",
        "user",
        "Which user is the admin user of the NeST"
            " (if unspecified, uses process uid)",
    },
    { Config_SystemUser,        
        "system_user",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "system",
        "user",
        "I don't know what this is",
    },
    { Config_AnonymousUser,     
        "anonymous_user",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "",
        "uid",
        "Which uid to use for anonymous users",
    },
    { Config_PasswordFile,      
        "passwd_filepath",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "/etc/passwd",
        "path",
        "Where NeST should read local user info",
    },
    { Config_GroupFile,         
        "group_filepath",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "/etc/group",
        "path",
        "Where NeST should read group user info",
    },
    { Config_X509CertDir,       
        "x509_cert_dir",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        ARG_REQUIRED,
        NULL,
        "path",
        "Where NeST should find the certicates dir for trusted CAs",
    },
    { Config_X509HostKey,       
        "x509_host_key",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        ARG_REQUIRED,
        NULL,
        "path",
        "Where NeST should find its gsi host key",
    },
    { Config_X509HostCert,       
        "x509_host_cert",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        ARG_REQUIRED,
        NULL,
        "path",
        "Where NeST should find its gsi host certificate",
    },
    { Config_GridMapFile,       
        "grid_mapfile_filepath",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "/etc/grid-security/grid-mapfile",
        "path",
        "Used as default list of allowed users, mapping to local uids",
    },
    { Config_MetaDir,           
        "meta_dir",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "meta-dir",
        "path",
        "Directory where NeST should write meta-data",
    },
    { Config_ActualGroup,       
        "actual_group",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "",
        "group",
        "I don't know what this is",
    },
    { Config_NestPort,          
        "nest_port",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "7878",
        "#",
        "Port on which NeST should listen for NeST protocol",
    },
    { Config_ProxyPort,         
        "proxy_port",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "7879",
        "#",
        "Port on which NeST proxies should listen for proxy communication",
    },
    { Config_HTTPPort,          
        "http_port",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "80",
        "#",
        "Port on which NeST proxies should listen for HTTP protocol",
    },
    { Config_FTPPort,           
        "ftp_port",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        ARG_REQUIRED,
        "21",
        "#",
        "Port on which NeST proxies should listen for FTP protocol",
    },
    { Config_GFTPPort,          
        "gftp_port",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "2811",
        "#",
        "Port on which NeST proxies should listen for Grid FTP protocol",
    },
    { Config_NFSPort,           
        "nfs_port",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        ARG_REQUIRED,
        "2049",
        "#",
        "Port on which NeST proxies should listen for NFS protocol",
    },
    { Config_DebugStderr,       
        "debug_to_stderr",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        NO_ARG_REQUIRED,
        NULL,
        NULL,
        "Set to make NeST debug to stderr"
    },
    { Config_CondorAdvertise,   
        "condor_advertise",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        ARG_REQUIRED,
        NULL,
        "path",
        "path to condor_advertise so that NeST can publish ad to matchmaker",
    },
    { Config_GlobusLocation,    
        "globus_location",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        ARG_REQUIRED,
        NULL,
        "path",
        "path to globus installation directory",
    },
    { Config_LogFile,           
        "logfile",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "/dev/stdout",
        "path",
        "where NeST should write its log messages",
    },
    { Config_DisableKernelQuota,
        "disable_kernel_quota",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        NO_ARG_REQUIRED,
        NULL,
        NULL,
        "forces NeST to do its own lot management without kernel support",
    }, 
    { Config_FdCacheSize,
        "fd_cachesize",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "25",
        "#",
        "how many fd's to cache in open",
    }, 
    { Config_FileCreateMode,
        "file_create_mode",
        NO_ARGUMENT,
        ARG_DEFINED,
        ARG_REQUIRED,
        "0600",
        "#",
        "Physical creation \"mode\" for user files",
    }, 
    { Config_DumpConfigFile,       
        "dump_config",
        NO_ARGUMENT,
        NO_ARG_DEFINED,
        NO_ARG_REQUIRED,
        NULL,
        NULL,
        "to dump a default config file to stdout",
    }, 
    { Config_PrintHelp,       
        "help",
        "h",
        NO_ARG_DEFINED,
        OPTIONAL_ARG,
        NULL,
        "[command]",
        "to print usage info or request detailed help on a particular flag",
    }, 
};

void
DebugConfig( const char *format, ... ) {
    if ( ! getenv( "NEST_DEBUG_CONFIG" ) ) {
        return;
    }
    va_list args;
    va_start( args, format );
    vfprintf( stderr, format, args );
    va_end( args );
}

bool
argMatch( const char *one, const char *two ) {
    if ( one == NO_ARGUMENT || two == NO_ARGUMENT ) return false;
    while( *one == '-' ) *one++;    // skip any flags
    while( *two == '-' ) *two++;    // skip any flags
    return ( ! strcmp( one, two ) );
}

bool
insertDefaultValues() {
    if ( CONFIG_VARIABLES_COUNT != NUMBER_OF_CONFIG_VARIABLES) {
        nest_die( "Problem with number of config vars in %s: %d != %d\n",
                __FILE__,
                CONFIG_VARIABLES_COUNT,
                NUMBER_OF_CONFIG_VARIABLES );
    }

    config_values = new struct config_values[CONFIG_VARIABLES_COUNT];

    for (int i = 0; i < (int) CONFIG_VARIABLES_COUNT; i++) {
        int which = config_variables[i].which_variable;
        config_values[which].argument= config_variables[i].name;
        config_values[which].defined = config_variables[i].defined; 
        config_values[which].value   = config_variables[i].default_value;
        config_values[which].short_arg=config_variables[i].short_arg;
        config_values[which].specified=NotSpecified;
        //DebugConfig( "Insert %s for %s.\n",
        //        config_values[which].value,
        //        config_values[which].argument );
    }

    return true;
}

bool
isFlag( const char *arg ) {
    return ( *arg == '-' );
}

struct config_variables *
findConfigVariable( const char *arg ) {
    for( int i = 0; i < (int) CONFIG_VARIABLES_COUNT; i++ ) {
        if ( argMatch( config_variables[i].name, arg ) ) {
            return &(config_variables[i]);
        } else if ( argMatch( config_variables[i].short_arg, arg ) ){
            return &(config_variables[i]);
        }
    }
    return NULL;
}

bool
takesArg( struct config_variables *cv, const char *possible_arg ) {
    switch( cv->requires_arg ) {
    case ARG_REQUIRED:      return true;
    case NO_ARG_REQUIRED:   return false;
    case OPTIONAL_ARG:      return ( possible_arg && !isFlag(possible_arg) );
    default:                assert( 0 ); return false;
    }
}

void
PrintHelp( const char *command ) {
    if ( command != NULL ) {
        struct config_variables *cv = findConfigVariable( command );
        if ( cv == NULL ) {
            printf( "No help is available for unknown command %s\n", command );
        } else {
            printf( "HELP %s\n"
                    "\tDefault value: %s\n"
                    "\tUsage:         %s %s\n"
                    "\tShortcut:      %s\n"
                    "\n"
                    "\tDescription:   %s\n",
                    command,
                    cv->default_value ? cv->default_value : "undefined",
                    command, 
                    cv->usage ? cv->usage : "",
                    cv->short_arg ? cv->short_arg : "no shortcut",
                    cv->help );
        }
    } else {
        printf( "Usage: %s\n", prog_name );
        for (int i = 0; i < (int) CONFIG_VARIABLES_COUNT; i++) {
            printf( "\t--%-25s %s\n", 
                    config_variables[i].name,
                    config_variables[i].usage );
        }
        printf( 
            "\n"
            "All arguments can also be stored in the config file (without the"
            " \'--\' prefix)\n"
            "For detailed information, type \'help <command>\'\n" 
            );
    }
}

void
DumpSampleConfigFile() {
    printf( "###############################################\n"
            "# Sample config file for %s\n"
            "# \tAll commands can also be passed on the command line with\n"
            "# \ta \'--\' prefix\n"
            "###############################################\n" 
            "\n\n",
            prog_name );

    for (int i = 0; i < (int) CONFIG_VARIABLES_COUNT; i++) {
        printf( "# %s\n"
                "%s%-25s %s\n",
                config_variables[i].help,
                config_variables[i].defined ? "" : "# ",
                config_variables[i].name,
                config_variables[i].default_value 
                    ? config_variables[i].default_value : ""
                );
        printf( "\n" );
    }

}

bool
ReadArgs( int argc, char **argv ) {
    prog_name = NestStrdup( argv[0] );

    for( int arg = 1; arg < argc; arg++ ) {

        if ( ! isFlag( argv[arg] ) ) {
            nest_die( "%s is an invalid command line arg.\n", argv[arg] );
        }

        struct config_variables *variable = findConfigVariable( argv[arg] );
        if ( variable == NULL ) {
            nest_die( "An unknown argument (%s) was specified\n", argv[arg] );
        } 

        const char *value = NULL;
        if( takesArg( variable, argv[arg+1] ) ) {
            if ( arg+1>= argc || isFlag( argv[arg+1] ) ) {
                nest_die( "Argument %s requires a value\n", argv[arg] );
            }
            value = NestStrdup( argv[arg+1] );
            arg++;  // skip the argument to the flag
        }
        StoreVariable( variable, value, SpecifiedCommandLine );
    }

    return true;
}

bool 
ReadConfigFile( const char *filename) {
    assert( filename );
    
    ifstream config_file(filename);
    if (!config_file) {
        if ( config_values[Config_ConfigFile].specified != NotSpecified ) {
            nest_die( "Can't find configuration file \"%s\"\n", filename);
        }
    } else {
        //config_file.flags(0);
        
        string line;
        int    line_number = 0;

        while (1) {
            // Note that because we are using getline and a string,
            // we don't have to worry about really long lines:
            // it will be handled for us.
            getline(config_file, line, '\n');
            line_number++;
            
            if (config_file.eof()) {
                break;
            } else {
                string variable, value;

                if (ExtractVariable(line, variable, value, line_number)) {
                    StoreVariable(variable.c_str(), value.c_str(), 
                                    filename, line_number);
                }
            }
            
        }
    }
    return true;
}

bool
argMatch( int which_var, const char *arg ) {
    if ( argMatch( config_values[which_var].argument, arg ) ) {
        return true;
    } else {
        return argMatch( config_values[which_var].short_arg, arg );
    }
}

const char *
findConfigFile( int argc, char **argv ) {

        // don't use a default value 
    const char *config_file = NULL; 

        // first see if it is specified in the environment
    if ( getenv( "NEST_CONFIG" ) ) {
        config_file = getenv( "NEST_CONFIG" );
        config_values[Config_ConfigFile].specified = SpecifiedEnvironment;
    }

        // then look through the args
    for( int i = 0; i < argc; i++ ) {
        if ( argMatch( Config_ConfigFile, argv[i] ) ) 
        {
            if ( i+1 >= argc ) {
                nest_die( "No argument provided to flag %s\n", argv[i]);
            }
            config_file = argv[i+1];
            config_values[Config_ConfigFile].specified = SpecifiedCommandLine;
        }
    }

    return config_file;
}

bool
ReadConfigFile( int argc, char **argv ) {
        // set up the initial config variables
    insertDefaultValues();

        // now read in values from the config file
    const char *config_file = findConfigFile( argc, argv );
    if ( config_file ) {
        DebugConfig( "Reading config file %s.\n", config_file );
        bool ret = ReadConfigFile( config_file );
        if ( ! ret ) return ret;
    } else {
        DebugConfig( "No config file found, using default values.\n" );
    }

        // weird special case for the actual config variable itself
    switch( config_values[Config_ConfigFile].specified ) {
    case SpecifiedConfigFile:
        nest_die( "Unable to specify a config file in the config file.\n" );
        break;
    case SpecifiedCommandLine:
            // reset it so when we read it again, won't complain about it
            // being multiply defined
        config_values[Config_ConfigFile].specified = NotSpecified;
        break;
    case NotSpecified:
    case SpecifiedEnvironment:
    default:
        break;
    }

        // then read in the command line args
    bool ret = ReadArgs( argc, argv );
    if ( ! ret ) return ret;

    DumpConfig();
    return ret;
}

void 
CleanupConfigFile(void)
{
    if (config_values != NULL) {
        for (int i = 0; i < (int) CONFIG_VARIABLES_COUNT; i++) {
            if (config_values[i].defined && config_values[i].value != NULL) {
                NestFree(config_values[i].value);
            }
        }
        delete [] config_values;
    }
    return;
}

bool
GetConfigValue(
    NestConfigVariable variable,
    const char **value)
{
    if ( config_values == NULL ) {
        nest_die( "Must call ReadConfigValue before GetConfigValue.\n" );
        return false;
    }
    
    bool is_defined = false;

    if (variable >= 0 
        && variable < (NestConfigVariable) CONFIG_VARIABLES_COUNT) {
        is_defined = config_values[variable].defined;
        if (value != NULL) {
            *value = config_values[variable].value;
        }
    }

    /*
    DebugConfig( "Lookup of %s is %s (%s)\n",
            config_values[variable].argument,
            ( is_defined ) ? "defined" : "undefined",
            ( is_defined ) ? config_values[variable].value : "UNDEF" );
    */
    return is_defined;
}

void DumpConfig(void)
{
    const char *value = NULL;
    for (int i = 0; i < (int) CONFIG_VARIABLES_COUNT; i++) {
        DebugConfig("%s is ", config_variables[i].name);
        if (GetConfigValue(config_variables[i].which_variable, &value)) {
            if (!value || NestStrlen(value) == 0) {
                DebugConfig("defined but empty.\n");
            } else {
                DebugConfig("\"%s\"\n", value);
            }
        } else {
            DebugConfig("(undefined)\n");
        }
    }
    return;
}

static void StoreVariable(
        struct config_variables *variable,
        const char *value,
        HowSpecified specified )
{
    DebugConfig( "Storing config value %s for variable %s.\n",
            value, variable->name );

    struct config_values *values = &config_values[variable->which_variable];
    values->defined = 1;
    values->value = (value) ? NestStrdup( value ) : NULL;
    if ( values->specified == specified ) {
        nest_die( "Config variable %s is multiply defined in the %s.\n",
                variable->name, 
                (specified == SpecifiedCommandLine) 
                    ? "command line" : "config file"
                );
    }
    values->specified = specified;
}

static void StoreVariable(
    const char *variable, 
    const char *value,
    const char *config_file,
    int        line_number)
{
    struct config_variables *cv = findConfigVariable( variable );
    if ( cv == NULL ) {
        nest_die( "An unknown configuration variable (%s) was "
                  "specified in config file %s on line %d.\n",
                  variable, 
                  config_file,
                  line_number);
    } else {
        StoreVariable( cv, value, SpecifiedConfigFile );
    }

    return;
}

static bool ExtractVariable(
    string &line,     // in, the line we read
    string &variable, // out, the name of the variable
    string &value,    // out, the value of the variable
    int    line_number)
{
    int i;
    int length;
    bool have_complete_line;

    i = 0;
    length = line.size();
    have_complete_line = false;

    variable = "";
    value = "";

    // Skip whitespace
    while (i < length && isspace(line[i])) {
        i++;
    }
    // Make sure we don't have an empty line or a comment
    if (i < length && line[i] != '#') {
        
        // First extract the variable name, which is the string
        // of characters up to the next whitespace.
        while (i < length && !isspace(line[i])) {
            variable += line[i];
            i++;
        }

        // Next, skip any whitespace
        while (i < length && isspace(line[i])) {
            i++;
        }

        // Finally, grab the variables value. 
        // If it is enclosed in quotes, we take everything between the quotes
        // If it isn't enclosed in quotes, we take up to the whitespace,
        // or a comment marker (#)
        if (i < length && line[i] == '"') {
            i++; // skip the quote marker
            while (i < length && line[i] != '"') {
                // Is this a backwhacked quote? Insert just the quote
                if (line[i] == '\\' && (i+1) < length && line[i+1] == '"') {
                    value += '"';
                    i += 2;
                } else {
                    value += line[i];
                    i++;
                }
            }
            if (line[i] != '"') {
                nest_warn( __FILE__, __FUNCTION__,
                        "Line %d of the configuration file is "
                        "missing a closing quote.\n", line_number);
            }
        } else {
            while (i < length && !isspace(line[i]) && line[i] != '#') {
                value += line[i];
                i++;
            }
        }
        if (variable.size() > 0) {
            have_complete_line = true;
        }
    }
    return have_complete_line;
}

    

