/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent,
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * For contact information, please visit:
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#ifndef     _NEST_FILE_CATALOG_H_
#define     _NEST_FILE_CATALOG_H_

#include "nest_common.h"

class NestFileCatalog {

    private:
        ClassAdCollection       *file_catalog;

        ClassAdParser           parser;
        ClassAdUnParser         unparser;
        LocalCollectionQuery    query;

        string                  root_dir;
        ViewName                file_catalog_view;

        NestReplyStatus get_physical_name(string &physical_name,
                                          string logical_name,
                                          string user,
                                          string group,
                                          string lot_handle);

        bool            is_access_allowed_for_entry(const char *file,
                                                    const char *user,
                                                    nest_access_rights_t flags);


    public:

        NestFileCatalog();
        ~NestFileCatalog();

        bool    is_file_present(const char *file);

        bool    is_system_file(const char *file);

        /*
            All 'file' names are represented from the respective user's
            logical root and are 'logical' names!
        */
        NestReplyStatus add_entry(const char *logical_file,
                                  const char *physical_file,
                                  const char *logical_dir,
                                  const char *user,
                                  const char *group,
                                  nest_obj_type_t file_type);

        /*
            For files and directories
        */
        NestReplyStatus add_entry(const char *file,
                                  const char *user,
                                  nest_obj_type_t type);

        /*
            For sym-links
        */
        NestReplyStatus add_entry(const char *newpath,
                                  const char *oldpath,
                                  const char *user,
                                  nest_obj_type_t type);

        NestReplyStatus delete_entry(const char *file);

        NestReplyStatus
        get_system_link_physical_name(char *physical_filename,
                                      const char *logical_filename);

        NestReplyStatus get_physical_name(char *physical_filename,
                                          const char *logical_filename);

        NestReplyStatus get_logical_name(char *logical_filename,
                                         const char *logical_filename);

        NestReplyStatus get_logical_name(char *logical_filename,
                                         const char *filename,
                                         const char *user);

        NestReplyStatus get_reverse_logical_name( char *dest,
                                                  const char *src );

        NestReplyStatus get_owner_user( char *user,
                                        const char *file );

        NestReplyStatus get_type(nest_obj_type_t *type,
                                 const char *file);

        NestReplyStatus get_size(off_t  *size,
                                 const char *file);

        NestReplyStatus set_size(const char *file,
                                 off_t  size);

        NestReplyStatus set_transfer_flag(const char *file,
                                          bool flag);

        NestReplyStatus get_acl(const char *file,
                                const char *name,
                                int is_user,
                                nest_access_rights_t *result);

        NestReplyStatus get_all_acl(NestACLList **result,
                                    const char *file,
                                    int is_user);

        NestReplyStatus set_acl(const char *file,
                                const char *name,
                                int is_user,
                                nest_access_rights_t argument);

        NestReplyStatus get_user_acl(const char *file,
                                     const char *name,
                                     nest_access_rights_t *result);

        NestReplyStatus set_user_acl(const char *file,
                                     const char *name,
                                     nest_access_rights_t argument);

        NestReplyStatus get_group_acl(const char *file,
                                      const char *name,
                                      nest_access_rights_t *result);

        NestReplyStatus set_group_acl(const char *file,
                                      const char *name,
                                      nest_access_rights_t argument);

        NestReplyStatus get_file_classad(const char *file,
                                         ClassAd** file_classad);

        NestReplyStatus get_file_uniquifier(const char *physical_filename,
                                            NestFilehandle* uniquifier );

        NestReplyStatus get_uid(user_id_t *uid,
                                const char *user);

        NestReplyStatus get_parent_dir(char *dest,
                                       const char *src);

        NestReplyStatus get_dest_file(char *dest,
                                      const char *src,
									  const char *user);

        bool    is_system_dir_access_allowed(const char *physical_dirname,
                                             user_id_t uid,
                                             user_id_t owner_uid,
                                             nest_access_rights_t flags);

        bool    is_system_file_access_allowed(const char *physical_filename,
                                              user_id_t uid,
                                              user_id_t owner_uid,
                                              nest_access_rights_t flags);

        bool    is_system_entry_access_allowed(const char *physical_filename,
                                               user_id_t uid,
                                               nest_access_rights_t flags);

        bool is_anonymous_access_allowed(const char *physical_filename,
                                         nest_access_rights_t flags);


        bool is_system_access_allowed(const char *file,
                                      const char *user,
                                      nest_access_rights_t flags);

        bool is_access_allowed(const char *file,
                               const char *user,
                               nest_access_rights_t flags);

        bool is_access_allowed_in_parent(const char *file,
                                         const char *user,
                                         nest_access_rights_t flags);

        bool is_dir_empty( const char *logical_dir );

        NestReplyStatus system_dir_list( NestFileInfo **result,
                                         const char *physical_dir );

        NestReplyStatus dir_list( NestFileInfo **result,
                                  const char *logical_dir );

        NestReplyStatus attach_lot_to_file( const char *logical_file,
                                            NestLotHandle *handle );

        NestReplyStatus detach_lot_from_file( const char *logical_file,
                                              NestLotHandle *handle );

        NestReplyStatus initialize(ClassAdCollection *resources,
                                   const char *rootdir);

        NestReplyStatus recover(ClassAdCollection *resources,
                                const char *rootdir);

        NestReplyStatus synchronize();
        NestReplyStatus shutdown();
};

#endif
