/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_filecache_model.h"
#include "nest_filecache_model_mgr.h"

//#define HASH_TABLE_SIZE 1037
//#define HASH_TABLE_SIZE 25717
#define HASH_TABLE_SIZE 184319
//#define HASH_TABLE_SIZE 205759

// constructor
nestFilecacheModel::nestFilecacheModel( ) {

}

// constructor
nestFilecacheModel::nestFilecacheModel( float megs, int type, 
        removeCallback removeCb, bool track_hits ) 
{
    
    this->size = this->avail = (long)(megs * 1024); 
    nest_debug( NEST_DEBUG_CACHEMODEL,
            "Init filecachemodel: %f MBs, %ld KBs.\n", megs, avail );
    this->tail = this->head  = NULL;
    this->entries = 0;
    this->type = type;
    this->removeCb = removeCb;

    if ( track_hits ) {
        this->infinite_cache = nest_hash_table_create( HASH_TABLE_SIZE, NULL );
        assert( infinite_cache );
        //cacheLog = nestStorageManager::openMetaFile( "cache.log",
        //        WRITE_FLAGS | O_TRUNC );
    } else {
        this->infinite_cache = NULL;
        this->cacheLog = NULL;
    }
    hits = compulsory_misses = evictions = capacity_misses = 0;

    log_misses = false;
    
    use_hashtable = true;
    if ( use_hashtable ) {
        this->ht = nest_hash_table_create( HASH_TABLE_SIZE, NULL );
        if ( ht == NULL ) {
            nest_debug( NEST_DEBUG_CACHEMODEL,  
                    "Unable to create hash table in cachemodel\n" );
            use_hashtable = false;
        }
    } else {
        this->ht = NULL;
    }
    debug();


}

void
nestFilecacheModel::logMiss( const char * file ) {
    struct timeval ts;
    NestGettimeofday( &ts );
    char buffer[NEST_MESSAGESIZE];
    NestSnprintf( buffer, NEST_MESSAGESIZE, 
            "<html><body>"
            "%ld %s"
            "<ul>", 
            ts.tv_sec, file );
    for ( nestFilecacheEntry *cursor = head; 
            cursor != NULL; 
            cursor = cursor->next ) 
    {
        NestSnprintf( &buffer[NestStrlen(buffer)], 
                NEST_MESSAGESIZE - NestStrlen(buffer),
                "<li>%s", cursor->filename );
    }
    NestSnprintf( &buffer[NestStrlen(buffer)], 
            NEST_MESSAGESIZE - NestStrlen(buffer),
            "</ul></body></html>" );
    //cacheLog->Sendn( buffer, NestStrlen( buffer ) );
}

void
nestFilecacheModel::writeAd( nestProtocol *ad ) {
    if ( infinite_cache != NULL ) {
        char buffer[NEST_MESSAGESIZE];
        NestSnprintf( buffer, NEST_MESSAGESIZE,
                "Hits         = %ld\n"
                "Evictions    = %ld\n"
                "Comp_Misses  = %ld\n"
                "Capac_Misses = %ld\n",
                hits, evictions, compulsory_misses, capacity_misses );
        ad->Sendn( buffer, NestStrlen( buffer ) );
    }
}
    
// returns whether some file which is no longer in the cache ever was and if so
// what its length was
bool 
nestFilecacheModel::rememberLength( const char *file, long *length ) {
    assert( infinite_cache != NULL );
    long *l = (long*)nest_hash_table_lookup( infinite_cache, file ); 
    if ( l != NULL ) {
        *length = *l;
        nest_debug( NEST_DEBUG_PROXY, "Remember length of %s: %ld (%ld)\n", 
                file, *l, *length );
        return true;
    }
    return false;
}

// returns whether file was already in cache
bool
nestFilecacheModel::remove( const char *file ) {

        // free it from infinite cache
    if ( infinite_cache != NULL )  {
        nest_debug( NEST_DEBUG_PROXY, 
                "Removing %s from infinite_cache model.\n", file );
        void *entry = nest_hash_table_lookup( infinite_cache, file );
        if ( entry != NULL ) {
            long *l = (long*)entry;
            NestFree( l );
        }
        nest_hash_table_remove( infinite_cache, file );
    }

        // make sure it is in here
    nestFilecacheEntry *entry = findEntry( file );
    if ( entry == NULL ) return false;

        // if we make it here, free the entry
        // don't use our removeCb, because it was the storage manager who
        // initiated this call, calling it again would recurse
        // the removeCb is only used for removals
        // initiated by ~this~ 
    removeEntry( entry );
    freeEntry( entry );
    return true;
}

float
nestFilecacheModel::hit_rate( ) {
    return (float)hits/(hits + capacity_misses + compulsory_misses);
}

long
nestFilecacheModel::getevictions() {
    return evictions;
}

long
nestFilecacheModel::gethits() {
    return hits;
}

long
nestFilecacheModel::getmisses() {
    return ( capacity_misses + compulsory_misses );
}

// returns whether file was already in cache
// this is called when tmp file is complete, so don't actually
// free the temp file name from the infinite cache...
bool
nestFilecacheModel::renameFile( const char *oldname, const char *newname ) {

    nestFilecacheEntry *entry = findEntry( oldname );
    if ( entry == NULL ) return false;

    if ( use_hashtable ) nest_hash_table_remove( ht, entry->filename );
    removeEntry( entry );
    NestFree( entry->filename );
    entry->filename = NestStrdup( newname );
    pushEntry( entry );
    if ( use_hashtable ) nest_hash_table_insert( ht, newname, (void *)entry );

    if ( infinite_cache && !nest_hash_table_lookup(infinite_cache, newname) ) {
        // don't free oldname from infinite_cache (see above)
        nest_debug( NEST_DEBUG_PROXY, "Renaming %s to %s in infinite cache.\n",
                oldname, newname );
        long *l = (long*)NestMalloc( sizeof(*l) );

            // TODO: need to check to make sure it was found.  possible
            // we just got formatted or otherwise deleted.
        *l = *(long*)nest_hash_table_lookup( infinite_cache, oldname ); 
        assert( l != NULL );
        nest_hash_table_insert( infinite_cache, newname, l );
        nest_debug( NEST_DEBUG_PROXY, "Remembering length %ld for %s.\n", 
                *l, newname );
    }
    return true;
}

NestReplyStatus
nestFilecacheModel::insert( bool *hit, const char *file, long filesize, 
                            int weight ) 
{

    long fsize = filesize / 1024;  // convert to KBs
    long needed = fsize; 
    nestFilecacheEntry *entry = NULL;

        // maybe it is already in the cache
    if ( NULL != ( entry = check_and_refresh( file ) ) ) {
        debug();
        if ( hit != NULL ) *hit = true;
        hits++;
            // check whether file is growing, 
            // if so need to fall through to make room for growth
            // also update value remembered in infinite cache
        if ( entry->size < fsize ) {
            needed -= entry->size;
            nest_debug( NEST_DEBUG_CACHEMODEL | NEST_DEBUG_PROXY,
                    "Growing cached file %s by %ld (old size %ld, new %ld).\n",
                    file, needed, entry->size, fsize );
            entry->size = fsize;
            if ( infinite_cache ) {
            long *rememberLength = (long*)nest_hash_table_lookup( 
                    infinite_cache, file );
            assert( rememberLength != NULL );
            *rememberLength = filesize; 
            }
        } else {
            return NEST_SUCCESS;
        }
    } else {
            // its a miss, but is it capacity or compulsory?
        if ( hit != NULL ) *hit = false;    
        if ( infinite_cache != NULL ) {
            if ( nest_hash_table_lookup( infinite_cache, file ) ) {
                capacity_misses++;
                if ( log_misses ) logMiss( file );
            } else {
                compulsory_misses++;
                long *l = (long*)NestMalloc(sizeof(*l));    // this leaks?
                *l = filesize;
                nest_hash_table_insert( infinite_cache, file, l );
                nest_debug( NEST_DEBUG_PROXY, 
                        "Remembering len %ld (0x%x) permanently for %s.\n",
                        *l, l, file );
            }
        } else {
                // we don't know whether it is capacity or compulsory
                // just assume capacity
            capacity_misses++;       
        }
    }


    if ( needed > size ) {
        nest_debug( NEST_DEBUG_CACHEMODEL,  
                "Attempt to insert %s which is too big (%ld > %ld)\n",
                file, needed, size );
        return NEST_INSUFFICIENT_SPACE;
    }
    
        // free old entries if necessary
    if ( needed > avail ) {
        long space_made = make_space( needed - avail );
        if ( needed > avail ) {
            nest_debug( NEST_DEBUG_CACHEMODEL | NEST_DEBUG_PROXY,
                    "Tried to make_space %ld, only made %ld.\n",
                    (long)(needed - avail ),
                    space_made );
            debug();
            validityCheck();
            return NEST_INSUFFICIENT_SPACE;
        }
    }

        // consume space for the inserted entry
    avail -= needed;
    if ( avail < 0 ) {
        assert( needed >= this->size );
        avail = 0;
        assert( avail == 0 );
    }

        // make an entry for it if it is a new insertion
    if ( entry == NULL ) {
        nestFilecacheEntry *entry = ( nestFilecacheEntry * )
            NestMalloc( sizeof ( nestFilecacheEntry ) );
        entry->size = fsize;
            // TODO: avoid string copy, maybe do the same in hash table
        entry->filename = NestStrdup( file );
        entry->reference = weight;
        entry->weight    = weight;
        entries++;

            // insert into hash and into list
        if ( use_hashtable ) nest_hash_table_insert( ht, file, (void *)entry );
        pushEntry( entry );
    }

    debug();
    return NEST_SUCCESS;
}
    
bool 
nestFilecacheModel::shrink( float megs ) {

        // this code is all screwed up.
        // lets just run with static size
        // the problem seems to be that it just grows and grows and grows
        // until it flips negative
    return true;
    
    long kbytes = long( megs * 1024 );
    this->size  -= kbytes ;
    long space_made = make_space( kbytes );

        // what do we do with avail when we shrink?
        // maybe we should just make it zero . . . if we just shrunk
        // the cache, then it is presumably full
    this->avail = 0;
    //this->avail -= space_made;
    if ( this->size <= 0 ) {
        nest_debug( NEST_DEBUG_CACHEMODEL,  
                "Tried to shrink filecache below zero.\n" );
        assert( avail < kbytes );
        validityCheck();
        this->size = this->avail = 0;
    }
    return ( (space_made > 0 ) ? true : false );
}

bool
nestFilecacheModel::grow( float megs ) {
        // this code is all screwed up.
        // lets just run with static size
    return true;
    
    long kbytes = long( megs * 1024 );
    this->size  += kbytes;
    this->avail += kbytes;
    return true;
}

float 
nestFilecacheModel::getsize( ) {
    if ( size == 0 ) {
        nest_warn( __FILE__, __FUNCTION__, "filecachesize has shrunk to zero.\n" );
        if ( avail != 0 && head != NULL ) {
            nest_warn( __FILE__, __FUNCTION__, "bad state in filecachemodel.\n" );
        }
    }
    return (float)size / 1024; 
}

long
nestFilecacheModel::getavail( ) {
    return avail;
}

nestFilecacheEntry *
nestFilecacheModel::pushEntry( nestFilecacheEntry *entry ) {

        // put it at the head
    if ( head != NULL ) {
        assert( head->prev == NULL );
        assert( tail->next == NULL );
        head->prev = entry;
        entry->next = head;
        head = entry;
        head->prev = NULL;
    } else {
            // insert into empty list
        assert( tail == NULL );
        head = tail = entry;
        entry->next = entry->prev = NULL;
    }
    return entry;
}

bool
nestFilecacheModel::check( const char *file ) {
    return ( findEntry( file ) != NULL );
}

nestFilecacheEntry *
nestFilecacheModel::check_and_refresh( const char *file ) {

    nestFilecacheEntry *entry = findEntry( file );
    if ( entry == NULL ) return NULL;

    //nest_debug( NEST_DEBUG_PROXY, "Refreshing %s in cache.\n", file );
    switch( this->type ) {
    case NEST_CACHE_CLOCK:
        entry->reference = entry->weight;
        break;
    case NEST_CACHE_LRU:
        removeEntry( entry );
        pushEntry( entry );
        break;
    default:
        assert( 0 );
    }

    return entry;
}

// this removes and returns an item from the list
nestFilecacheEntry *
nestFilecacheModel::removeEntry( nestFilecacheEntry *entry ) {

    nestFilecacheEntry *next = entry->next;
    nestFilecacheEntry *prev = entry->prev;

    if ( prev != NULL && next != NULL ) {
            // cursor is safely in middle of list
        prev->next = next;
        next->prev = prev;
    } else if ( prev != NULL ) {
            // next must be null, so entry must be old tail
        assert( entry == tail );    // TODO: while running BLAST on proxys
                                    // this once failed . . . 
        tail = prev;
        tail->next = NULL;
    } else {
            // prev must be null, so entry must be old head
            // it might also be old tail
        assert( entry == head );
        if ( next != NULL ) {
            head = next;
            head->prev = NULL;
        } else {
                // removing the last entry
            assert( entry == tail );
            head = tail = NULL;
        }
    }

    return prev;
}

long
nestFilecacheModel::make_space( long needed ) {

    static nestFilecacheEntry *cursor = tail;
    nestFilecacheEntry *tmp;
    long space_made = 0;
    int cycles = 0;

        // this is for clock
    if ( type == NEST_CACHE_CLOCK ) {
        while( space_made < needed ) {
            if ( cursor == NULL ) {
                    // we've moved off the end of the list
                cursor = tail;
                if ( head == NULL ) {
                        // empty list
                    assert( tail == NULL );
                    break;
                }
                    // need to cycle more than max weight of an entry
                    // proxys assign different weights to
                    // different cached blocks depending on how
                    // important they are
                    // written blocks are infinitely important
                    // primary copies have initial reference count of two
                    // secondary copies have init  reference count of one
                cycles++;
                if ( cycles >= 4 ) break;
            }
            assert( cursor != NULL );

            if (cursor->weight != INFINITE_IMPORTANCE && cursor->reference == 0)
            {
                    // remove this entry
                tmp = cursor;
                NestReplyStatus status = NEST_SUCCESS;
                if ( removeCb != NULL ) {
                        // see if the storage manager will allow this file
                        // to be removed
                    status = removeCb( tmp->filename );  
                    if ( status == NEST_SUCCESS ) {
                        evictions++;
                    }
                } 
                if ( status == NEST_SUCCESS ) {
                        // we can free this entry 
                    evictions++;
                    cursor = removeEntry( cursor );
                    space_made += freeEntry( tmp );
                } else {
                        // we can't free this entry,
                    cursor = cursor->prev;
                }
            } else {
                    // decrement the reference count and keep looking
                if ( cursor->weight != INFINITE_IMPORTANCE ) {
                        // allow possibility to give some files
                        // extra preference
                    cursor->reference--;
                    assert( cursor->reference >= 0 );
                }
                cursor = cursor->prev;
            }
        }
        goto SPACE_MADE;
    }

        // this is for lru
    cursor = tail;
    while( space_made < needed ) {
            // free the tail, unless tail has already been removed
        if ( tail == NULL ) {
            assert( head == NULL );
            break;
        }

        NestReplyStatus status = NEST_SUCCESS;

            // move up to next entry
        nestFilecacheEntry *stale = cursor;
        cursor = cursor->prev;

            // see whether storage manager will allow this file to be removed
        if ( removeCb != NULL ) {
                // the removeCb will call removeEntry and freeEntry for us
            status = removeCb( stale->filename );
            if ( status == NEST_SUCCESS ) {
                evictions++;
            }
        }

            // actually free old entry if removeCb was successful 
        if ( status == NEST_SUCCESS ) {
            evictions++;
            removeEntry( stale );
            space_made += freeEntry( stale );
        }

        if ( cursor == NULL ) {
                // we've moved off the top of the list, nothing left to free
            break;
        }
    }

SPACE_MADE:
    return space_made;
}

NestReplyStatus
nestFilecacheModel::format( ) {
        // this wipes everything out that isn't locked
    nestFilecacheEntry *cursor = tail;
    NestReplyStatus ret = NEST_SUCCESS;
    NestReplyStatus removeStatus = NEST_SUCCESS;
    while ( cursor != NULL ) {
        nestFilecacheEntry *erase = cursor;
        cursor = cursor->prev;
        if ( removeCb != NULL ) {
                // actually ask storage to delete the physical files 
            removeStatus = removeCb( erase->filename );
            if ( removeStatus != NEST_SUCCESS ) ret = removeStatus;
        }
        if ( removeStatus == NEST_SUCCESS ) {
                // free our state
            removeEntry( erase );
            freeEntry( erase );
        }
    }

        // wipe out infinite cache
    if ( infinite_cache ) {
        nest_hash_table_delete( infinite_cache );
        infinite_cache = nest_hash_table_create( HASH_TABLE_SIZE, NULL );
        assert( infinite_cache );
    }
    return ret; 
}

bool
nestFilecacheModel::validityCheck() {
    if ( entries <= 0 || head == NULL ) {
        assert( entries == 0 );
        assert( head == NULL );
        assert( tail == NULL );
        //assert( avail == size );
        if ( avail != size ) {
                // this will happen after a file bigger than cache sweeps
                // everything out, and it itself doesn't fit
            nest_warn( __FILE__, __FUNCTION__, "nothing in cache, but avail (%ld) != size (%ld)\n",
                avail, size );
            avail = size;
        }
        /*
           // this may be too slow and it has never failed anyway
    } else {
            // count the entries by hand, make sure it matches our count
        nestFilecacheEntry *cursor = head;
        int i = 1;
        while ( ( cursor = cursor->next ) != NULL ) { i++; }
        assert( i == entries );
        */
    }
    return true;
}

long
nestFilecacheModel::freeEntry( nestFilecacheEntry *entry ) {
    long entry_size = entry->size;
    avail += entry_size;
    if ( use_hashtable ) nest_hash_table_remove( ht, entry->filename );
    NestFree( entry->filename );
    NestFree( entry );
    entries--;
    validityCheck();
    return entry_size;
}

long
nestFilecacheModel::getFilesize( const char *file ) {
    nestFilecacheEntry *entry = findEntry( file );
    return ( entry ) ? entry->size * 1024 : -1;
}

void
nestFilecacheModel::setFilesize( const char *file, long size ) {
    nestFilecacheEntry *entry = findEntry( file );
    if ( entry ) entry->size = size / 1024;     // KBs
}

nestFilecacheEntry *
nestFilecacheModel::findEntry( const char *file ) {
    nestFilecacheEntry *entry = NULL; 

    if ( use_hashtable ) {
        entry = (nestFilecacheEntry *) 
            nest_hash_table_lookup( ht, file );
    } else {
            // scan linked list looking for entry
        entry = head;
        while ( entry != NULL ) {
            if ( strcmp( file, entry->filename ) == 0 ) {
                break;
            }
            entry = entry->next;
        }
    }

        // size should never be 0 anyway
    if ( entry != NULL ) assert( size != 0 );
    return entry;
}

void
nestFilecacheModel::debug( ) {

    nest_debug( NEST_DEBUG_CACHEMODEL, 
            "Filecache model: size %ld, avail %ld, entries %ld\n", 
            size, avail, entries );
    nestFilecacheEntry *entry = head;
    while ( entry != NULL ) {
        nest_debug( NEST_DEBUG_CACHEMODEL, 
                "\t%s %ld %d\n", entry->filename, entry->size, 
                entry->reference );
        entry = entry->next;
    }
}
