/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#ifndef _NEST_LOT_MANAGER_H_

#define _NEST_LOT_MANAGER_H_


#include "nest_quota_manager.h"
#include "nest_device_manager.h"

#include "nest_common.h"

#define RMCurrentList   ".RMCurrentList"
#define RMPendingList   ".RMPendingList"

#define SCHED_NEXT      0

class NestLotManager {

    private:
        ClassAdCollection       *lot_catalog;
        ClassAdCollection       *identities;

        ViewName                lot_catalog_view;

        ClassAdParser           parser;
        ClassAdUnParser         unparser;
        LocalCollectionQuery    query;

        rsv_id_t                next_rsv_id;

        char                    root_dir[NEST_MAXPATHLEN];
        double                  storage_space;
        double                  max_lot_duration;
        bool                    has_kernel_quota;

        QuotaManager            *QuotaMngr;
        DeviceManager           *DeviceMngr;

        rsv_id_t                get_next_rsv_id();

        NestReplyStatus         sync_next_rsv_id();

        NestReplyStatus         request_lot(NestLotHandle *handle,
                                            NestLotHandle *parent_handle,
                                            const char *parent_name,
                                            const char *name,
                                            int is_user,
                                            double size,
                                            double duration,
                                            int lot_type=G_LOT,
		                    			    int cost = 1);

        NestReplyStatus         list_sub_lots(NestLotList **result,
                                              NestLotHandle *handle);
  
        NestReplyStatus         list_lots(NestLotList **result,
                                                  const char *name,
                                                  int is_user);

        NestReplyStatus         LotClassAd_to_NestLot(ClassAd *lot_ad,
                                                      NestLot *rstat);

        NestReplyStatus         get_actual_name(char *actual_name,
                                                const char *perc_name,
                                                int is_user);

        NestReplyStatus         get_lot_and_share(const char *arg,
                                                  char *lot,
                                                  double *share);

        NestReplyStatus         change_lot_usage(const char *lot_key,
                                                 double *size,
                                                 bool allocate_as_much = false);

        NestReplyStatus         add_lot_to_file(const char * lot_handle,
                                                double size,
                                                const char * file_key);

        NestReplyStatus         extend_current_lot(const char * file_key,
                                                   double size);

        NestReplyStatus         check_default_lot(const char * file_key,
                                                  double size,
                                                  const char * user);

        NestReplyStatus         reclaim_lot_space( const char *file_key,
                                                   double size );

    public:

        NestLotManager(const char *rootDir,
                       double storage_space,
                       double max_lot_duration);
        ~NestLotManager();

        bool kernel_quota_enabled();
        double delete_dead_rsvs(double req_blks,
                                int is_user,
                                bool aggressive = true);

        NestReplyStatus update_lot_usage(const char *logical_filename,
                                         double size);

        NestReplyStatus update_file_lot_usage( const char *logical_file,
                                               double size );

        NestReplyStatus update_dir_lot_usage( const char *logical_file,
                                              double size );

        NestReplyStatus update_user_lot_usage( const char *logical_file,
                                               double size );
        
        NestReplyStatus update_group_lot_usage( const char *logical_file,
                                                double size );

        NestReplyStatus modify_lot_usage(const char *user,
                                         const char *logical_filename,
                                         NestLotHandle  *from_lot,
                                         double from_share,
                                         NestLotHandle  *to_lot);

        NestReplyStatus set_user_default_lot( NestLotHandle *handle,
                                              const char *user);

        NestReplyStatus request_user_lot(NestLotHandle *handle,
                                         const char *user,
                                         double size,
                                         double duration,
                    					 const char *lot_user,
                                         const char *parent_user,
                                         const char *parent_group,
                                         NestLotHandle *parent_handle,
                                         int lot_type = G_LOT,
					                     int cost = 1);

        NestReplyStatus request_group_lot(NestLotHandle *handle,
                                          const char *user,
                                          double size,
                                          double duration,
                                          const char *lot_group,
                                          const char *parent_user,
                                          const char *parent_group,
                                          NestLotHandle *parent_handle,
                    					  int lot_type = G_LOT,
                                          int cost = 1);
        
        NestReplyStatus    get_parent_lot(const char *lot_user,
                                          const char *parent_user,
                                          NestLotHandle *parent_handle,
                                          double size,
                                          bool is_user);
   
        bool    is_access_allowed_to_request(const char *lot_user,
                                             char *parent_user,      
                                             NestLotHandle *parent_handle);
   
        bool    is_access_allowed_to_lot(NestLotHandle *handle,
                                         const char *user,
                                         int is_user = IS_USER,
                                         bool allow_group_members = true,
                                         bool allow_admin = true);

        bool    is_access_allowed_to_user(const char *user,
                                          const char *lot_user,
                                          int is_user = IS_USER,
                                          bool allow_group_members = true,
                                          bool allow_admin = true);

        NestReplyStatus list_user_lots(NestLotList **result,
                                       const char *user,
                                       const char *lot_user);

        NestReplyStatus list_group_lots(NestLotList **result,
                                        const char *req_user,
                                        const char *group);

        NestReplyStatus update_lot(NestLotHandle *handle,
                                   const char *user,
                                   double more_size,
                                   double more_duration,
                                   int more_cost,
                                   bool update_total_size = true);

        NestReplyStatus change_lot_type(NestLotHandle *handle,
                                        const char *user,
                                        int lot_type,
                                        double more_duration);

        NestReplyStatus get_lot_status(NestLot *result,
                                       NestLotHandle *handle);

        NestReplyStatus get_lot_status(NestLot *result,
                                       NestLotHandle *handle,
                                       const char *user); 

        NestReplyStatus terminate_lot(NestLotHandle *handle,
                                      const char *user,
                                      int sched_next = SCHED_NEXT);

        double free_space ( int is_user );
        double total_space( int is_user );
        double available_lot_space(NestLotHandle *handle);
        double available_lot_space(const char *user,
                                   const char *logical_file);
        double available_file_lot_space(const char *logical_file);
        double available_dir_lot_space(const char *logical_file);
        double available_user_lot_space(const char *user);
        double available_group_lot_space(const char *user);
        NestReplyStatus available_group_lot_space(const char *user,
                                                  const char *group,
                                                  double *size);

        NestReplyStatus add_user_to_lot(NestLotHandle *handle,
                                        const char *user,
                                        const char *lot_user);
        NestReplyStatus delete_user_from_lot(NestLotHandle *handle,
                                             const char *user,
                                             const char *lot_user);

        NestReplyStatus get_file_usage(const char *logical_filename,
                                       double *usage);

        int init();
        NestReplyStatus initialize(ClassAdCollection *resources);
        NestReplyStatus recover(ClassAdCollection *resources);

        NestReplyStatus synchronize();
        NestReplyStatus shutdown();
};


#endif
