/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nfs_prot.h"
#include "nfs_mount.h"
#include "nfs_util.h"
#include "nest_client.h"

#include <stdio.h>
#include <stdlib.h>
#include <rpc/pmap_clnt.h>
#include <string.h>
#include <memory.h>
#include <sys/socket.h>
#include <netinet/in.h>

#ifndef SIG_PF
#define SIG_PF void(*)(int)
#endif

#include "nest_protocol_nfs.h"
#include "nest_pthread_util.h"
#include "nfs_prot.h"
#include "nfs_mount.h"
#include "nfs_util.h"

typedef union {
        // mount args
    dirpath mountproc_mnt_1_arg;
    dirpath mountproc_umnt_1_arg;

        // nfs args
    nfs_fh nfsproc_getattr_2_arg;
    sattrargs nfsproc_setattr_2_arg;
    diropargs nfsproc_lookup_2_arg;
    nfs_fh nfsproc_readlink_2_arg;
    readargs nfsproc_read_2_arg;
    writeargs nfsproc_write_2_arg;
    createargs nfsproc_create_2_arg;
    diropargs nfsproc_remove_2_arg;
    renameargs nfsproc_rename_2_arg;
    linkargs nfsproc_link_2_arg;
    symlinkargs nfsproc_symlink_2_arg;
    createargs nfsproc_mkdir_2_arg;
    diropargs nfsproc_rmdir_2_arg;
    readdirargs nfsproc_readdir_2_arg;
    nfs_fh nfsproc_statfs_2_arg;
} nfs_argument;

typedef union {
        // mount replies
    fhstatus    fhstatus_reply;     // mountproc_mnt_1
    mountlist   mountlist_reply;    // mountproc_dump_1
    exports     exports_reply;      // mountproc_export_1, exportall

        // nfs replies
    attrstat    attrstat_reply;     // getattr, setattr, write
    diropres    diropres_reply;     // lookup, create, mkdir
    readlinkres readlinkres_reply;  // readlink
    readres     readres_reply;      // read
    nfsstat     nfsstat_reply;      // remove, rename, link, symlink, rmdir
    readdirres  readdirres_reply;   // readdir
    statfsres   statfsres_reply;    // statfs

        // common replies
    char        *other_reply;       // umnt_1, umntall_1, mount_null
                                    // nfsproc_null, nfsproc_root, writecache

} nfs_reply;

// when a new rpc packet is ready, the main server will wake from its select
// loop because we merge the rpc readfds into the main set
//
// the main server will then call my_svc_getreqset which calls the appropriate
// rpc handler with the demarshalled arguments.  the handlers then create
// nest_request objects which are then assigned to a new nest_client object
// which is queued
//
// after this happens, the nest_server will then ask each protocol with a
// ready socket to accept any new connections.  dequeue and return.




enum {
    MOUNT_REQUEST,
    NFS_REQUEST,
};

int                  nestProtocolNFS::pipes[2];

// this queue doesn't need to be threadsafe as it is accessed by only a single
// thread running in accept
queue<nestClient *>  nestProtocolNFS::clients;

extern struct xp_ops my_svcudp_op;

nestClient *
nestProtocolNFS::accept( int socket, fd_set *readys, int accepts ) {

        // would like to call my_svc_getreqset here but problem is that 
        // we can only do it once per select
        // currently nest server keeps calling accept until it returns
        // NULL . .. 
        // only call it if this is the first time we're called this select loop
        // first receive any rpc calls, then return the clients
    if ( accepts == 0 ) {
        //nest_debug( NEST_DEBUG_NFS, "Receiving any rpc callbacks.\n" );
        my_svc_getreqset( request_handler, readys );
        //nest_debug( NEST_DEBUG_NFS, "Done receiving rpc callbacks.\n" );
    }

    nestClient *client = NULL;
    if ( ! clients.empty() ) {
        client = (nestClient *)clients.front();
        clients.pop();
    }
    /*
    if ( ! clients.empty() ) {
        nest_warn( __FILE__, __FUNCTION__, "More than one NFS client accepted in main select loop.\n");
    }
    */
    if ( client != NULL ) {
        nest_debug( NEST_DEBUG_NFS, "Accepting an nfs client: %d\n", 
                client->getId() );
    }
    return client;
}


// TODO new connection must somehow enqueue 
// accept will then do the corresponding dequeue

void
nestProtocolNFS::newconnection( nestRequest *request,
                                SVCXPRT     *transp, 
                                xdrproc_t   xdr_result,
                                xdrproc_t   xdr_argument,
                                caddr_t     argument,
                                svc_req     *rqstp,
                                int         general_type ) 
{
    nfs_fh *fh;
    char   *path;
    int uid;
    gid_t gid;
    struct authunix_parms *unix_cred = NULL;

        // cache the last uid lookup
    static int last_uid = -1;

    static char uname[NEST_MAXPATHLEN];    
    
    nest_debug( NEST_DEBUG_NFS, "NFS request successfully parsed, "
            "creating a new nfs connection\n" );
    assert( request != NULL );

    nestProtocolNFS *connection = new nestProtocolNFS( 
        transp, xdr_result, xdr_argument, argument, general_type,
        rqstp->rq_proc );

    nestClient *client = new nestClient( 
            (nestProtocol*)connection,  svc_getcaller( rqstp->rq_xprt ) );

        // if the request passed a file handle, 
        // turn it back into a path
        // this path should either be the cwd for lookups
        // or the target file for other operations
        // if the reverselookup fails, then the file handle has 
        // gone stale and the client must refresh it
    if ( ( fh = (nfs_fh*)request->getFilehandle() ) != NULL ) {
        path = reverseLookup( (nfs_fh*)request->getFilehandle() );
        if ( path != NULL ) {
            if (general_type == NFS_REQUEST 
                 && (    rqstp->rq_proc == NFSPROC_LOOKUP 
                      || rqstp->rq_proc == NFSPROC_MKDIR 
                      || rqstp->rq_proc == NFSPROC_READDIR ) ) 
            {
                client->setCwd( path );
            } else {
                if ( request->getNestFile() == NULL ) {
                    request->setNestFile( path );
                }
            }
        } else {
            nest_warn( __FILE__, __FUNCTION__,"Rev lookup of %d failed.  Returning stale fh error.\n",
                nfsFilehandleToInt( fh ) );
            NestFree( request );
            request = new nestRequest( BAD_REQUEST, 0, 0, 0, NEST_STALE_FH, 0 );
        }
        request->clearFilehandle();
    }

        // check authentication
    switch( rqstp->rq_cred.oa_flavor ) {
    case AUTH_UNIX:
        unix_cred = (struct authunix_parms *)rqstp->rq_clntcred;
        uid = unix_cred->aup_uid;
        gid = unix_cred->aup_gid;
        if ( uid == last_uid ) {
            nest_debug( NEST_DEBUG_NFS, "Using cached uid->name (%d:%s)\n",
                    uid, uname );
            client->setName( uname );
        } else {
            if ( NestGet_uname( uname, uid ) < 0 ) {
                nest_warn( __FILE__, __FUNCTION__,"Uid %d is unknown.  Returning error.\n",
                    uid ); 
                NestFree( request );
                request = new nestRequest( BAD_REQUEST, 0, 0, 0, 
                        NEST_UNKNOWN_USER, 0 );
            } else {
                client->setName( uname );
                last_uid = uid;
            }
        }
        nest_debug( NEST_DEBUG_NFS, "NFS request used AUTH_UNIX (%d.%d->%s)\n",
                uid, gid, client->getIdentity() );
        break;
    case AUTH_DES:
        nest_debug( NEST_DEBUG_NFS, "NFS request used AUTH_DES\n" ); 
        break;
    case AUTH_NONE:
        nest_debug( NEST_DEBUG_NFS, "NFS request used AUTH_NONE\n" ); 
        break;
    case AUTH_KERB:
        nest_debug( NEST_DEBUG_NFS, "NFS request used AUTH_KERB\n" ); 
        break;
    default:
        assert( 0 );
    }

    client->storeRequest( request );
    clients.push( client );
}

char *
nestProtocolNFS::reverseLookup( nfs_fh *fh ) {
    NestFilehandle nest_fh;
    memcpy( &nest_fh, fh, sizeof( NestFilehandle ) );
    return nestStorageManager::filehandleToPath( &nest_fh ); 
}



void
nestProtocolNFS::mount_request_handler(
        struct svc_req *rqstp, register SVCXPRT *transp)
{
    nfs_argument *argument = (nfs_argument*)NestMalloc( sizeof( *argument));

    nestRequest *request;
    xdrproc_t _xdr_argument, _xdr_result;
    nestRequest *(*local)(char *, struct svc_req *);

    nfs_dump_request( rqstp, mount_request( rqstp->rq_proc ) ); 

    switch (rqstp->rq_proc) {
    case MOUNTPROC_NULL:
        _xdr_argument = (xdrproc_t) xdr_void;
        _xdr_result = (xdrproc_t) xdr_void;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            mountproc_null_1_svc;
        break;

    case MOUNTPROC_MNT:
        _xdr_argument = (xdrproc_t) xdr_dirpath;
        _xdr_result = (xdrproc_t) xdr_fhstatus;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            mountproc_mnt_1_svc;
        break;

    case MOUNTPROC_DUMP:
        _xdr_argument = (xdrproc_t) xdr_void;
        _xdr_result = (xdrproc_t) xdr_mountlist;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            mountproc_dump_1_svc;
        break;

    case MOUNTPROC_UMNT:
        _xdr_argument = (xdrproc_t) xdr_dirpath;
        _xdr_result = (xdrproc_t) xdr_void;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            mountproc_umnt_1_svc;
        break;

    case MOUNTPROC_UMNTALL:
        _xdr_argument = (xdrproc_t) xdr_void;
        _xdr_result = (xdrproc_t) xdr_void;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            mountproc_umntall_1_svc;
        break;

    case MOUNTPROC_EXPORT:
        _xdr_argument = (xdrproc_t) xdr_void;
        _xdr_result = (xdrproc_t) xdr_exports;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            mountproc_export_1_svc;
        break;

    case MOUNTPROC_EXPORTALL:
        _xdr_argument = (xdrproc_t) xdr_void;
        _xdr_result = (xdrproc_t) xdr_exports;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            mountproc_exportall_1_svc;
        break;

    default:
        svcerr_noproc (transp);
        return;
    }
    memset ((char *)argument, 0, sizeof (*argument));

        // parse args, if failure send error message
    if (!svc_getargs (transp, _xdr_argument, (caddr_t) argument)) {
        svcerr_decode (transp);
        NestFree( argument );
        return;
    }

        // get a nest request object
        // TODO create a client object and assign the req to it
        // TODO queue the client and return it when asked to accept
    request = (*local)((char *)argument, rqstp);
    if ( request != NULL ) {
        newconnection(  request,
                        transp, 
                        _xdr_result, 
                        _xdr_argument, 
                        (caddr_t) argument,
                        rqstp,
                        MOUNT_REQUEST );
    }

    /*
    if (result != NULL && !svc_sendreply(transp, _xdr_result, result)) {
        svcerr_systemerr (transp);
    }
    */

    /*
        // free args - this is now done in the nfs_protocol destructor
    if (!svc_freeargs (transp, _xdr_argument, (caddr_t) argument)) {
        nest_warn( __FILE__, __FUNCTION__,  "unable to free arguments");
        exit (1);
    }
    */

    return;
}

void
nestProtocolNFS::request_handler( 
        struct svc_req *rqstp, register SVCXPRT *transp )
{
    nest_debug( NEST_DEBUG_NFS, "incoming nfs request.\n" );
    if ( rqstp->rq_prog == MOUNTPROG ) {
        mount_request_handler( rqstp, transp );
    } else {
        assert( rqstp->rq_prog == NFS_PROGRAM );
        nfs_request_handler( rqstp, transp );
    }
}

void
nestProtocolNFS::nfs_request_handler( 
        struct svc_req *rqstp, register SVCXPRT *transp)
{
    nfs_argument *argument = (nfs_argument *) NestMalloc( sizeof( *argument ) );

    nestRequest *request;
    xdrproc_t _xdr_argument, _xdr_result;
    nestRequest *(*local)(char *, struct svc_req *);

    nfs_dump_request( rqstp, nfs_request( rqstp->rq_proc ) ); 

    switch (rqstp->rq_proc) {
    case NFSPROC_NULL:
        _xdr_argument = (xdrproc_t) xdr_void;
        _xdr_result = (xdrproc_t) xdr_void;
        local = (nestRequest *(*)(char *, struct svc_req *)) nfsproc_null_2_svc;
        break;

    case NFSPROC_GETATTR:
        _xdr_argument = (xdrproc_t) xdr_nfs_fh;
        _xdr_result = (xdrproc_t) xdr_attrstat;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            nfsproc_getattr_2_svc;
        break;

    case NFSPROC_SETATTR:
        _xdr_argument = (xdrproc_t) xdr_sattrargs;
        _xdr_result = (xdrproc_t) xdr_attrstat;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            nfsproc_setattr_2_svc;
        break;

    case NFSPROC_ROOT:
        _xdr_argument = (xdrproc_t) xdr_void;
        _xdr_result = (xdrproc_t) xdr_void;
        local = (nestRequest *(*)(char *, struct svc_req *)) nfsproc_root_2_svc;
        break;

    case NFSPROC_LOOKUP:
        _xdr_argument = (xdrproc_t) xdr_diropargs;
        _xdr_result = (xdrproc_t) xdr_diropres;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            nfsproc_lookup_2_svc;
        break;

    case NFSPROC_READLINK:
        _xdr_argument = (xdrproc_t) xdr_nfs_fh;
        _xdr_result = (xdrproc_t) xdr_readlinkres;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            nfsproc_readlink_2_svc;
        break;

    case NFSPROC_READ:
        _xdr_argument = (xdrproc_t) xdr_readargs;
        _xdr_result = (xdrproc_t) xdr_readres;
        local = (nestRequest *(*)(char *, struct svc_req *)) nfsproc_read_2_svc;
        break;

    case NFSPROC_WRITECACHE:
        _xdr_argument = (xdrproc_t) xdr_void;
        _xdr_result = (xdrproc_t) xdr_void;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            nfsproc_writecache_2_svc;
        break;

    case NFSPROC_WRITE:
        _xdr_argument = (xdrproc_t) xdr_writeargs;
        _xdr_result = (xdrproc_t) xdr_attrstat;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            nfsproc_write_2_svc;
        break;

    case NFSPROC_CREATE:
        _xdr_argument = (xdrproc_t) xdr_createargs;
        _xdr_result = (xdrproc_t) xdr_diropres;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            nfsproc_create_2_svc;
        break;

    case NFSPROC_REMOVE:
        _xdr_argument = (xdrproc_t) xdr_diropargs;
        _xdr_result = (xdrproc_t) xdr_nfsstat;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            nfsproc_remove_2_svc;
        break;

    case NFSPROC_RENAME:
        _xdr_argument = (xdrproc_t) xdr_renameargs;
        _xdr_result = (xdrproc_t) xdr_nfsstat;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            nfsproc_rename_2_svc;
        break;

    case NFSPROC_LINK:
        _xdr_argument = (xdrproc_t) xdr_linkargs;
        _xdr_result = (xdrproc_t) xdr_nfsstat;
        local = (nestRequest *(*)(char *, struct svc_req *)) nfsproc_link_2_svc;
        break;

    case NFSPROC_SYMLINK:
        _xdr_argument = (xdrproc_t) xdr_symlinkargs;
        _xdr_result = (xdrproc_t) xdr_nfsstat;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            nfsproc_symlink_2_svc;
        break;

    case NFSPROC_MKDIR:
        _xdr_argument = (xdrproc_t) xdr_createargs;
        _xdr_result = (xdrproc_t) xdr_diropres;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            nfsproc_mkdir_2_svc;
        break;

    case NFSPROC_RMDIR:
        _xdr_argument = (xdrproc_t) xdr_diropargs;
        _xdr_result = (xdrproc_t) xdr_nfsstat;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            nfsproc_rmdir_2_svc;
        break;

    case NFSPROC_READDIR:
        _xdr_argument = (xdrproc_t) xdr_readdirargs;
        _xdr_result = (xdrproc_t) xdr_readdirres;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            nfsproc_readdir_2_svc;
        break;

    case NFSPROC_STATFS:
        _xdr_argument = (xdrproc_t) xdr_nfs_fh;
        _xdr_result = (xdrproc_t) xdr_statfsres;
        local = (nestRequest *(*)(char *, struct svc_req *)) 
            nfsproc_statfs_2_svc;
        break;

    default:
        svcerr_noproc (transp);
        return;
    }
    memset ((char *)argument, 0, sizeof (*argument));

        // try to read the args, send an error if parse fails
    if (!svc_getargs (transp, _xdr_argument, (caddr_t) argument)) {
        svcerr_decode (transp);
        return;
    }

        // ask handle to turn args into a request object,
        // all of the handlers are in nfs_request_handler
        // TODO create a client object and assign the req to it
        // TODO queue the client and return it when asked to accept
    request = (*local)((char *)argument, rqstp);
    if ( request != NULL ) {
        newconnection(  request,
                        transp, 
                        _xdr_result, 
                        _xdr_argument, 
                        (caddr_t) argument,
                        rqstp,
                        NFS_REQUEST );
    }

    return;
}

int 
nestProtocolNFS::listen( int port ) {

        // register our services
    register SVCXPRT *transp;
    char *err_msg = NULL;


    // for some reason, only one nfs client works at a time
    // I tried doing multiple registers but that didn't work
    //for( int i = 0; i < 2; i++ ) {

        // start listening to mount requests
    pmap_unset (MOUNTPROG, MOUNTVERS);

    transp = svcudp_create(RPC_ANYSOCK);
    if (transp == NULL) {
        err_msg = "cannot create udp service.";
        goto NFS_LISTEN_DONE;
    }
    if (!svc_register(transp, MOUNTPROG, MOUNTVERS, request_handler, 
                IPPROTO_UDP)) {
        err_msg = "unable to register (MOUNTPROG, MOUNTVERS, udp).";
        goto NFS_LISTEN_DONE;
    }

    my_svcudp_op = *(const xp_ops *)(transp->xp_ops);

    /*
       // don't listen anymore to tcp bec. muthian's library is udp only
    transp = svctcp_create(RPC_ANYSOCK, 0, 0);
    if (transp == NULL) {
        err_msg = "cannot create tcp service.";
        goto NFS_LISTEN_DONE;
    }
    if (!svc_register(transp, MOUNTPROG, MOUNTVERS, request_handler, 
                IPPROTO_TCP)) {
        err_msg =  "unable to register (MOUNTPROG, MOUNTVERS, tcp).";
        goto NFS_LISTEN_DONE;
    }
    */

        // start listening to nfs requests
    pmap_unset (NFS_PROGRAM, NFS_VERSION);

    transp = svcudp_create(RPC_ANYSOCK);
    if (transp == NULL) {
        err_msg =  "cannot create udp service.";
        goto NFS_LISTEN_DONE;
    }
    if (!svc_register(transp, NFS_PROGRAM, NFS_VERSION, request_handler, 
                IPPROTO_UDP)) 
    {
        err_msg =  "unable to register (NFS_PROGRAM, NFS_VERSION, udp).";
        goto NFS_LISTEN_DONE;
    }

    /*
       // don't listen anymore to tcp bec. muthian's library is udp only
    transp = svctcp_create(RPC_ANYSOCK, 0, 0);
    if (transp == NULL) {
        err_msg =  "cannot create tcp service.";
        goto NFS_LISTEN_DONE;
    }
    if (!svc_register(transp, NFS_PROGRAM, NFS_VERSION, request_handler, 
                IPPROTO_TCP)) {
        err_msg =  "unable to register (NFS_PROGRAM, NFS_VERSION, tcp).";
        goto NFS_LISTEN_DONE;
    }
    */

    // don't do the my_svc_run here, add the svc_fdset to the main select loop
    /*
    my_svc_run ();
    nest_warn( __FILE__, __FUNCTION__,  "my_svc_run returned");
    exit (1);
    */
    /* NOTREACHED */

NFS_LISTEN_DONE:
    if ( err_msg != NULL ) {
        nest_die( "NFS listen failure: \n\t%s.\n"
                "\tDid you run portmap?\n", err_msg);
    }

        // I don't think this will be necessary
        // use a pipe to communicate with the main admin thread
    if ( NestSpipe( pipes ) < 0 ) {
        nest_perror( "s_pipe" );
        return -1;
    }
    return pipes[1];
}

nestProtocolNFS::nestProtocolNFS(   SVCXPRT *transp, 
                                    xdrproc_t xdr_result,
                                    xdrproc_t xdr_argument,
                                    caddr_t   argument,
                                    int       general_type,
                                    int       req_type
                                    ) 
{
    this->transp        = transp;
    this->xdr_result    = xdr_result;
    this->xdr_argument  = xdr_argument;
    this->argument      = argument;
    this->gen_type      = general_type;
    this->req_type      = req_type;
}

nestProtocolNFS::~nestProtocolNFS( ) {
    if (!svc_freeargs (transp, xdr_argument, argument)) {
        nest_warn( __FILE__, __FUNCTION__,  "unable to free arguments");
    }
    NestFree( argument );
    my_svcudp_destroy( transp, 0 );
    nest_debug( NEST_DEBUG_NFS, "Removed nfs protocol connection.\n" );
}

bool
nestProtocolNFS::persistent( ) {
    return false;
}

int
nestProtocolNFS::getSocket() {
    return( -1 );
}

// I'd like to return true but currently we use the hasSocket() test to 
// determine whether we can work in POP and NOB and NFS can't
bool
nestProtocolNFS::hasSocket() {
    return false;
}

ssize_t
nestProtocolNFS::Send( const char *buffer, ssize_t n ) {

    NestReplyStatus status = NEST_SUCCESS;

    nest_debug( NEST_DEBUG_NFS, "nfs sending %d bytes of data\n", n );
    assert( gen_type == NFS_REQUEST && req_type == NFSPROC_READ );

    nfs_reply reply;
    reply.readres_reply.status = NFS_OK;
    reply.readres_reply.readres_u.reply.data.data_len = n;
    reply.readres_reply.readres_u.reply.data.data_val = buffer;
    setFileAttributes( &reply.readres_reply.readres_u.reply.attributes, NULL );
    int maxtries = 5, tries;
    for ( tries = 0; tries < maxtries; tries++ ) {
        status = sendReply( (char*)&reply );
        if ( status == NEST_SUCCESS ) break;
    }
    if ( status == NEST_SUCCESS && tries > 0 ) {
        nest_warn( __FILE__, __FUNCTION__, "Took %d tries to send nfs data.\n", tries );
    }
    return ( ( status == NEST_SUCCESS ) ? n : -1 );
}

ssize_t
nestProtocolNFS::Receive( char *buffer, ssize_t n ) {
    
    nest_debug( NEST_DEBUG_NFS, "nfs receiving %d bytes of data\n", n );

    int size_to_copy = (u_int) n < ((writeargs*)argument)->data.data_len ?
                       (u_int) n : ((writeargs*)argument)->data.data_len;
    strncpy(buffer, ((writeargs*)argument)->data.data_val, size_to_copy);
    return size_to_copy;
}

NestReplyStatus
nestProtocolNFS::sendFilehandle( NestFileInfo *fileinfo ) {

    NestFilehandle *handle = &(fileinfo->uniquifier);
    nfs_reply   reply;

        // check that we are replying to the appropriate request
    assert(    ( gen_type == MOUNT_REQUEST && req_type == MOUNTPROC_MNT ) 
            || ( gen_type == NFS_REQUEST   && req_type == NFSPROC_LOOKUP ) );
    
    nest_debug( NEST_DEBUG_NFS, "Sending filehandle: %d\n", 
            filehandleToInt( handle ) );

    if ( gen_type == MOUNT_REQUEST ) {
        reply.fhstatus_reply.fhs_status = NFS_OK;   
        filehandleCopy( &(reply.fhstatus_reply.fhstatus_u.fhs_fhandle), handle);
    } else {
        reply.diropres_reply.status = NFS_OK;
        filehandleCopy( &(reply.diropres_reply.diropres_u.diropres.file.data), 
                handle );
        setFileAttributes( 
                &(reply.diropres_reply.diropres_u.diropres.attributes),
                fileinfo );
    }

    return sendReply( (char*)&reply );
}

NestReplyStatus
nestProtocolNFS::sendFileinfo( NestFileInfo *info ) {
    nfs_reply reply;

        // check that we are replying to the appropriate request
    assert( gen_type == NFS_REQUEST && req_type == NFSPROC_GETATTR );

    reply.attrstat_reply.status = NFS_OK; 
    setFileAttributes( &(reply.attrstat_reply.attrstat_u.attributes), info ); 

    nest_debug( NEST_DEBUG_NFS, "nfs sending file info\n" );
    return sendReply( (char*)&reply );
}

NestReplyStatus
nestProtocolNFS::setFileattr( NestFileInfo *info, const char *fname, 
                              void *fileattr) {

    struct sattrargs *fattr = (struct sattrargs*)fileattr;
    NestChmod( fname, fattr->attributes.mode );
    NestChown( fname, fattr->attributes.uid, 
                               fattr->attributes.gid );
    
    nfs_reply reply;

        // check that we are replying to the appropriate request
    assert( gen_type == NFS_REQUEST && req_type == NFSPROC_SETATTR );

    reply.attrstat_reply.status = NFS_OK; 
    setFileAttributes( &(reply.attrstat_reply.attrstat_u.attributes), info ); 

    nest_debug( NEST_DEBUG_NFS, "nfs setting file attributes\n" );
    return sendReply( (char*)&reply );

}


bool
nestProtocolNFS::sendDiskspace( double avail, double total ) {
    nfs_reply reply;

        // check that we are replying to the appropriate request
    assert( gen_type == NFS_REQUEST && req_type == NFSPROC_STATFS );

    reply.statfsres_reply.status = NFS_OK; 

    reply.statfsres_reply.statfsres_u.reply.tsize  = (u_int) 8192;
    reply.statfsres_reply.statfsres_u.reply.bsize  = (u_int) 4096;
    reply.statfsres_reply.statfsres_u.reply.blocks = (u_int) total / 4096;
    reply.statfsres_reply.statfsres_u.reply.bfree  = (u_int) avail / 4096;
    reply.statfsres_reply.statfsres_u.reply.bavail = (u_int) avail / 4096;
    
    nest_debug( NEST_DEBUG_NFS, "nfs sending statfs info\n" );
    return sendReply( (char*)&reply );
}

char *
lastComponent( char *path ) {
        // first of all, if it is a directory, make sure to remove
        // the slash as the final character
    if ( path[strlen(path)-1] == '/' ) path[strlen(path)-1] = '\0';
    
    char *file = strrchr( path, '/' );
    return ( file ) ? ++file : path;
}


bool
nestProtocolNFS::sendDirectoryList( NestFileInfo *list ) {
    nfs_reply reply;

    reply.readdirres_reply.status = NFS_OK;
    reply.readdirres_reply.readdirres_u.reply.eof = true;     
    reply.readdirres_reply.readdirres_u.reply.entries = NULL;

    int offset = 1;
    while ( list != NULL ) {
        entry *e = (entry *)NestMalloc( sizeof( entry ) );     

        memcpy( e->cookie, &offset, NFS_COOKIESIZE );
        e->fileid     = (u_int)list->uniquifier;
        e->name       = lastComponent(list->name);
        e->nextentry  = reply.readdirres_reply.readdirres_u.reply.entries;
        offset++;

        reply.readdirres_reply.readdirres_u.reply.entries = e;
        list = list->next;
    }
    nest_debug( NEST_DEBUG_NFS, "nfs sending directory list\n" );
    return sendReply( (char*)&reply );
}

NestReplyStatus
nestProtocolNFS::sendReply( char *result ) {
    NestReplyStatus status;

    lockMutex();
    
    if ( ! svc_sendreply( transp, xdr_result, result ) ) {
        nest_warn( __FILE__, __FUNCTION__, "nfs send reply failure\n" );
        svcerr_systemerr( transp );
        status = NEST_NO_CONNECTION;
    } else {
        status = NEST_SUCCESS; 
    }

    unlockMutex();
    
    return status;
}

void
nestProtocolNFS::lockMutex( ) {
}

void
nestProtocolNFS::unlockMutex( ) {
}

bool
nestProtocolNFS::sendFilesize( long size ) {
    assert( 0 );
    return false;
}

bool
nestProtocolNFS::sendErrorMessage( const char *filename, const char *func_name,  NestReplyStatus nest_err ) {

    nfs_reply reply;

    enum nfsstat nfs_err = NestStatusToNFSStatus( nest_err );
    //assert( nfs_err != NFS_OK );

    nest_debug( NEST_DEBUG_NFS, "%s:%s:send nfs error: %s\n", filename, func_name, nfs_error(nfs_err) ); 

    if ( gen_type == MOUNT_REQUEST ) {

        switch( req_type ) {
        case MOUNTPROC_MNT:        
        case MOUNTPROC_UMNT:     
            reply.fhstatus_reply.fhs_status = nfs_err;
            break;
        case MOUNTPROC_UMNTALL: 
        case MOUNTPROC_EXPORT: 
        case MOUNTPROC_EXPORTALL:   
        case MOUNTPROC_NULL:        
        case MOUNTPROC_DUMP:      
        default:                    assert( 0 );
        }
    } else {
        assert( gen_type == NFS_REQUEST );

        switch( req_type ) {
        case NFSPROC_REMOVE:        
        case NFSPROC_RENAME:        
        case NFSPROC_LINK:          
        case NFSPROC_SYMLINK:       
        case NFSPROC_RMDIR:         
            reply.nfsstat_reply = nfs_err;
            break;
            
        case NFSPROC_LOOKUP:        
        case NFSPROC_MKDIR:         
        case NFSPROC_CREATE:        
            reply.diropres_reply.status = nfs_err;
            if ( nest_err == NEST_SUCCESS && req_type == NFSPROC_MKDIR ) {
                nest_debug( NEST_DEBUG_NFS, "Sending ack for mkdir\n" );
                setFileAttributes( 
                    &(reply.diropres_reply.diropres_u.diropres.attributes), 
                    NULL );
            }
            break;

        case NFSPROC_GETATTR:       
        case NFSPROC_SETATTR:       
        case NFSPROC_WRITE:         
            reply.attrstat_reply.status = nfs_err;
            break;

        case NFSPROC_READLINK:      
            reply.readlinkres_reply.status = nfs_err;
            break;

        case NFSPROC_READ:          
            reply.readres_reply.status = nfs_err;
            break;

        case NFSPROC_READDIR:       
            reply.readdirres_reply.status = nfs_err;
            break;

        case NFSPROC_STATFS:        
            reply.statfsres_reply.status = nfs_err;
            break;

        case NFSPROC_NULL:          
        case NFSPROC_ROOT:          
        case NFSPROC_WRITECACHE:    
        default:                    assert( 0 );
        }
    }

    return ( sendReply( (char*)&reply ) == NEST_SUCCESS );
}

bool
nestProtocolNFS::sendAck(  ) {
    return sendErrorMessage( __FILE__, __FUNCTION__,  NEST_SUCCESS );
}

bool
nestProtocolNFS::sendFinalAck( int transferType ) {
        // I think there is nothing to do here
    nest_debug( NEST_DEBUG_NFS, "nfs noop for sendFinalAck\n" );
    return true;
}

NestReplyStatus
nestProtocolNFS::setFileAttributes( fattr *attributes, NestFileInfo *info ) {

    if ( info != NULL ) {
        nest_debug( NEST_DEBUG_NFS, "nfs setting file attributes.\n" );

        attributes->size = (unsigned) info->size;

        switch( info->type ) {
        case NEST_DIR:
            attributes->size = 4096;
            attributes->type = NFDIR; 
            attributes->mode = 0040000; 
            nest_debug( NEST_DEBUG_NFS, "nfs setting file attributes dir.\n" );
            break;
        case NEST_LINK:     
            attributes->type = NFLNK; 
            attributes->mode = 0120000;
            break;
        case NEST_FILE:     
            attributes->type = NFREG; 
            attributes->mode = 0100000;
            break;
        default:
            nest_warn( __FILE__, __FUNCTION__, "Unknown filetype in nfs setFileAttributes.\n" );
            memset( attributes, 0, sizeof( fattr ) );
            return NEST_NOT_YET_IMPLEMENTED;
        }

            // lie about access permissions
        attributes->mode |= 0777;

        attributes->uid = attributes->gid = 0;
        attributes->nlink = 1;
        attributes->rdev = 1;
        attributes->fsid = 1;
        attributes->fileid = (u_int)info->uniquifier;

        attributes->blocksize = 4096;
        attributes->blocks = (attributes->size / attributes->blocksize) + 1;

            // make up times.  This particular time is for 5/29/2002
        attributes->atime.seconds  = attributes->mtime.seconds = 
            attributes->ctime.seconds  = 1022715085;
        attributes->atime.useconds = attributes->mtime.useconds = 
            attributes->ctime.useconds = 0;

        return NEST_SUCCESS;
    } else {
        nest_debug( NEST_DEBUG_NFS, "nfs Not setting file attributes.\n" );
        memset( attributes, 0, sizeof( fattr ) );
        return NEST_NOT_YET_IMPLEMENTED;
    }
}

bool
nestProtocolNFS::sendTransferInfo( nestRequest *req ) {

        // handle special case where the I/O request is for
        // zero length . . .
    if ( req->getStart() >= req->getEnd() ) {
        assert( req->getStart() == req->getEnd() );
        nfs_reply reply;
        reply.readres_reply.status = NFS_OK;
        reply.readres_reply.readres_u.reply.data.data_len = 0;
        reply.readres_reply.readres_u.reply.data.data_val = NULL;
        setFileAttributes( &reply.readres_reply.readres_u.reply.attributes, 
                NULL );
        return ( sendReply( (char*)&reply ) == NEST_SUCCESS );
    } else {
        nest_debug( NEST_DEBUG_NFS, "nfs noop for sendTransferInfo\n" );
        return true;
    }
}

nestRequest* 
nestProtocolNFS::receiveRequest( fd_set *ready ){
    nest_debug( NEST_DEBUG_NFS, "nfs receive request returning close.\n" );
    return new nestRequest( CLOSE_CONNECTION_REQUEST, 0,0,0,0,0 );
//    nest_debug( NEST_DEBUG_NFS, "nfs receive request returning null.\n" );
//    return NULL;
}
