/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent,
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * For contact information, please visit:
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_request.h"
#include <string.h>

nestRequest::nestRequest( const nestRequest *req ) {
	init( req->type,
		 req->nestfilename,
		 req->arg1,
		 req->arg2,
		 req->start,
		 req->end,
		 req->arg3,
		 req->arg4,
		 req->arg5,
         req->arg6 );
}

// different requests pass different semantic data for arg1 and arg2
// quota requests pass the user name for arg1
// third party transfers pass the remotehost as arg1 and remote file as arg2
nestRequest::nestRequest(
						 int type,
                         char *nestfile,
                         char *arg1,
                         char *arg2,
                         long start,
                         long end,
                         char *arg3,
                         char *arg4,
                         char *arg5,
                         char *arg6 )
{
	init( type, nestfile, arg1, arg2, start, end, arg3, arg4, arg5, arg6);
}

void
nestRequest::init(		 int type,
                         char *nestfile,
                         char *arg1,
                         char *arg2,
                         long start,
                         long end,
                         char *arg3,
                         char *arg4,
                         char *arg5,
                         char *arg6 )
{
    this->type     = type;
    this->end      = end;
    this->start    = start;
    this->chunk    = 0;

    this->nestfilename = this->arg1 = this->arg2 = this->arg3 = NULL;
    this->arg4 = this->arg5 = this->arg6 = NULL;

    if ( nestfile != NULL && NestStrlen( nestfile ) ) {
        this->nestfilename = NestStrdup( nestfile );
    } 

    if ( arg1 != NULL && NestStrlen( arg1 ) ) {
        this->arg1 = NestStrdup( arg1 );
    }

    if ( arg2 != NULL && NestStrlen( arg2 ) ) {
        this->arg2 = NestStrdup( arg2 );
    }

    if ( arg3 != NULL && NestStrlen( arg3 ) ) {
        this->arg3 = NestStrdup( arg3 );
    }

    if ( arg4 != NULL && NestStrlen( arg4 ) ) {
        this->arg4 = NestStrdup( arg4 );
    }

    if ( arg5 != NULL && NestStrlen( arg5 ) ) {
        this->arg5 = NestStrdup( arg5 );
    }

    if ( arg6 != NULL && NestStrlen( arg6 ) ) {
        this->arg6 = NestStrdup( arg6 );
    }

    this->fh = NULL;
    this->status = NEST_NOT_YET_INITIALIZED;

    NestGettimeofday( &this->createtime );
}

nestRequest::~nestRequest( ) {
    //NestGettimeofday( &this->endtime );
    if ( nestfilename ) NestFree ( nestfilename );
    if ( arg1 )         NestFree ( arg1 );
    if ( arg2 )         NestFree ( arg2 );
    if ( arg3 )         NestFree ( arg3 );
    if ( arg4 )         NestFree ( arg4 ); 
    if ( arg5 )         NestFree ( arg5 ); 
    if ( arg6 )         NestFree ( arg6 );
}

long
nestRequest::getAge() {
    return NestElapsedTime( &createtime );
}

void
nestRequest::setFilehandle( void *fh ) {
    this->fh = fh;
}

void
nestRequest::clearFilehandle( ) {
    this->fh = NULL;
}

void *
nestRequest::getFilehandle( ) {
    return fh;
}

void
nestRequest::setFileattr( void *fattr ) {
    this->fattr = fattr;
}

void
nestRequest::clearFileattr( ) {
    this->fattr = NULL;
}

void *
nestRequest::getFileattr( ) {
    return fattr;
}

char *
nestRequest::toString() {
    switch( type ) {
    case TIME_REQUEST:                       // 0
        return "time";
    case CLIENT_SEND_FILE:
        return "client send";
    case CLIENT_RECV_FILE:
        return "client recv";
    case FILE_DELETE_REQUEST:
        return "rm";
    case BANDWIDTH_STATISTICS_REQUEST:
        return "statistics";
    case DIR_LIST_REQUEST:                   // 5
        return "ls";
    case MKDIR_REQUEST:
        return "mkdir";
    case CWD_REQUEST:
        return "cwd";
    case PWD_REQUEST:
        return "pwd";
    case FILESIZE_REQUEST:
        return "filesize";
    case SERVER_TERMINATE_REQUEST:           // 10
        return "server terminate";
    case CLIENT_TERMINATE_REQUEST:
        return "client terminate";
    case TEST_MESSAGE:
        return "test message";
    case TEST_CLIENT_TO_SERVER:
        return "test client-server";
    case TEST_SERVER_TO_CLIENT:
        return "test server-client";
    case REQUEST_DONE:                       // 15
        return "request done";
    case CHANGE_USER_REQUEST:
        return "change user";
    case AVAILABLE_SPACE_QUERY:
        return "available";
    case FORMAT_DISK_REQUEST:
        return "format";
    case CONNECTION_REQUEST:
        return "open connection";
    case CLOSE_CONNECTION_REQUEST:           // 20
        return "close connection";
    case THIRD_PARTY_GET:
        return "3p get";
    case THIRD_PARTY_PUT:
        return "3p put";
    case DEBUG_REQUEST:
        return "debug";
    case RMDIR_REQUEST:
        return "rmdir";
    case NAME_COLLECTION:                    // 25
        return "name collection";
    case UNNAME_COLLECTION:
        return "unname collection";
    case QUERY_COLLECTION:
        return "query collection";
    case HELP_REQUEST:
        return "help";
    case RHELP_REQUEST:
        return "rhelp";
    case NEST_SYNC:                          // 30
        return "sync";
    case REQUEST_NEST_AD:
        return "ad";
    case LINK_REQUEST:
        return "link";
    case ADD_NEST_USER:
        return "add user";
    case ADD_NEST_GROUP:
        return "add group";
    case DELETE_NEST_USER:                   // 35
        return "delete user";
    case BAD_REQUEST:
        return "bad request";
    case DELETE_NEST_GROUP:
        return "delete group";
    case ADD_USER_TO_GROUP:
        return "add user to group";
    case DELETE_USER_FROM_GROUP:
        return "delete user from group";
    case REQUEST_USER_LOT:                   // 40
        return "request user lot";
    case REQUEST_GROUP_LOT:
        return "request group lot";
    case LIST_USER_LOTS:
        return "list user lots";
    case LIST_GROUP_LOTS:
        return "list group lots";
    case GET_LOT_STAT:
        return "get lot";
    case UPDATE_LOT:                         // 45
        return "update lot";
    case CHANGE_LOT_T:
        return "chage lot type";
    case TERMINATE_LOT:
        return "terminate lot";
    case VERSION_REQUEST:
        return "version";
    case PHYSNAME_REQUEST:
        return "physname";
    case GET_ACL:                            // 50
        return "get acl";
    case SET_ACL:                            
        return "set acl";
    case GET_GROUP_ACL:
        return "get group acl";
    case SET_GROUP_ACL:
        return "set group acl";
    case LOT_MOVE_FILE:
        return "move lot";
    case GET_FILEHANDLE:                     // 55   
        return "filehandle";
    case FILEINFO_REQUEST:                   
        return "fileinfo";
    case PROXY_GET:
        return "proxy get";
    case PROXY_PUT:
        return "proxy put";
    case PROXY_REGISTER:
        return "proxy register";
    case PROXY_QUERY:                        // 60
        return "proxy query";
    case COPY_REQUEST:                        
        return "copy";
    case ADD_USER_TO_LOT:
        return "add user to lot";
    case DELETE_USER_FROM_LOT:
        return "delete user from lot";
    case FILE_RENAME_REQUEST:
        return "mv";
    case ATTACH_LOT_TO_FILE:
        return "attach lot to file";
    case DETACH_LOT_FROM_FILE:
        return "detach lot from file";
	case PROXY_INDEX_PUSH:
		return "proxy index push";
	case PROXY_FLUSH:
		return "proxy flush";
    case SETATTR_REQUEST:
        return "set fileattr";
	case SET_COLLECTION_MTIME:
		return "set collection time";
	case SET_COLLECTION_VERSION:
		return "set collection version";
	case SET_COLLECTION_COMMENT:
		return "set collection comment";
	case GET_COLLECTION_MTIME:
		return "get collection time";
	case GET_COLLECTION_VERSION:
		return "get collection version";
	case GET_COLLECTION_COMMENT:
		return "get collection comment";
    default:
    	return "unknown request type : __FILE__ __FUNCTION__ __LINE__"; 
    }
}

int
nestRequest::getValue() {
    return (int)start;
}

int
nestRequest::getType( ) {
    return type;
}

bool
nestRequest::isWrite( int type ) {
	return ( type == CLIENT_SEND_FILE );
}

bool
nestRequest::isWrite() {
	return( isWrite( type ) );
}

bool
nestRequest::isThirdParty( ) {
    return ( type == THIRD_PARTY_GET || type == THIRD_PARTY_PUT );
}

long
nestRequest::getStart( ) {
    return start;
}

long
nestRequest::getEnd( ) {
    return end;
}

long
nestRequest::getChunk( ) {
    return chunk;
}

void
nestRequest::setChunk( long c ) {
    chunk = c;
}

int
nestRequest::getAuthType( ) {
    return getValue();
}

NestReplyStatus
nestRequest::getErrorCode( ) {
    return ((NestReplyStatus) getValue());
}

void
nestRequest::setEnd( long end ) {
    this->end = end;
}

char *
nestRequest::getProtocol( ) {
    return arg3;
}

/*
void
nestRequest::setQuota( long quota ) {
    this->end = quota;
}
*/

void
nestRequest::setStart( long start ) {
    this->start = start;
}

const char *
nestRequest::getNestFile( ) {
    return nestfilename;
}

const char *
nestRequest::getOldNestFile( ) {
    return arg1;
}

const char *
nestRequest::getNewNestFile( ) {
    return arg2;
}

/*
void
nestRequest::restoreNestFileName() {
    if ( nestfilename ) NestFree( nestfilename );
    nestfilename = oldnestfilename;
    oldnestfilename = NULL;
}

const char *
nestRequest::getOldNestFile() {
    return oldnestfilename;
}

void
nestRequest::renameNestFile( char *file ) {
    if ( oldnestfilename ) NestFree( oldnestfilename );
    oldnestfilename = nestfilename;
    nestfilename = file;
}

long
nestRequest::getOldStart() {
    return oldstart;
}
*/

const char *
nestRequest::getProxyLocation( ) {
    return nestfilename;
}

NestReplyStatus
nestRequest::setNestFile( const char *file ) {
    assert( nestfilename == NULL );
    nestfilename = NestStrdup( file );
    return NEST_SUCCESS;
}

char *
nestRequest::getDebugValue( ) {
    return nestfilename;
}

const char *
nestRequest::getCollectionName( ) {
    return arg1;
}

time_t
nestRequest::getCollectionMtime( ) {
    time_t tmp;
    if(!arg2){
        return -1;
    }
    sscanf(arg2, "%ld", &tmp);
    return tmp;
}

const char *
nestRequest::getCollectionVersion( ) {
    return arg2;
}

const char *
nestRequest::getCollectionComment( ) {
    return arg2;
}

char *
nestRequest::getLinkName( ) {
    return arg1;
}

char *
nestRequest::getRemoteFile( ) {
    return arg2;
}

char *
nestRequest::getRemoteHost( ) {
    return arg1;
}

const char *
nestRequest::getGroup( ) {
    return arg2;
}

const char *
nestRequest::getUser( ) {
    return NestGetModifiedIdentity( arg1 );
}

const char*
nestRequest::getIdentity( ){
    return arg3;
}

const char *
nestRequest::getTargetUser( ) {
    return NestGetModifiedIdentity( arg1 );
}

bool
nestRequest::getAclIsGroup( ) {
    if ( ! arg3 ) return false;
    else return ( ! strcasecmp( arg3, "true" ) );
}

const char *
nestRequest::getTargetAcl( ) {
    return arg2;
}

int
nestRequest::getLotType( ) {
    int tmp;
    if(!arg2){
        return -1;
    }
    sscanf(arg2, "%d", &tmp);
    return tmp;
}

char*
nestRequest::getLotUser( ) {
    if(arg3 && !strcmp(arg3, "0"))
        return NULL;
    return NestGetModifiedIdentity( arg3 );
}

char*
nestRequest::getParentLotUser( ) {
    if(arg4 && !strcmp(arg4, "0"))
        return NULL;
    return NestGetModifiedIdentity( arg4 );
}

char*
nestRequest::getParentLotGroup( ) {
    if(arg5 && !strcmp(arg5, "0"))
        return NULL;
    return NestGetModifiedIdentity( arg5 );
}
    
rsv_id_t
nestRequest::getParentLot( ){
    rsv_id_t rsv_id;
    if(arg6)
        sscanf(arg6, "%lu" , &rsv_id);
    return rsv_id; 
}

double
nestRequest::getLotSize( ) {
    double tmp;
    if (!arg1) {
        return 0;
    }
    sscanf(arg1,"%lf", &tmp);
    return tmp;
}

double
nestRequest::getLotDuration( ) {
    double tmp;
    if (!nestfilename) {
        return 0;
    }
    sscanf(nestfilename,"%lf",&tmp);
    return tmp;
}

int
nestRequest::getLotHandle(NestLotHandle *handle) {
    if (!handle || !this->arg2) {
        return -1;
    }

    rsv_id_t    tmp_id;

    sscanf(this->arg2,"%ld", &tmp_id);

    NestAdd_id_to_handle(handle, tmp_id);

    return 0;
}

int
nestRequest::getSourceLotHandle(NestLotHandle *handle) {
    if (!handle || !this->arg1) {
        return -1;
    }

    rsv_id_t    tmp_id;

    sscanf(this->arg1,"%ld", &tmp_id);

    NestAdd_id_to_handle(handle, tmp_id);

    return 0;
}

int
nestRequest::getDestLotHandle(NestLotHandle *handle) {
    if (!handle || !this->arg2) {
        return -1;
    }

    rsv_id_t    tmp_id;

    sscanf(this->arg2,"%ld", &tmp_id);

    NestAdd_id_to_handle(handle, tmp_id);

    return 0;
}

double
nestRequest::getLotMoveBytes( ) {
    double tmp;
    if (!nestfilename) {
        return 0;
    }
    sscanf(arg3, "%lf",  &tmp);
    return tmp;
}

int
nestRequest::getNewLotType( ) {
    int lot_type = -1;

    if(!arg1) {
        return -1;
    }
    sscanf(arg1, "%d", &lot_type);
    return lot_type;
}
