
#include "nfs_util.h"

void 
nfs_dump_request( struct svc_req *rqstp, char *req ) {
    char *ip = NestDns_lookup( svc_getcaller( rqstp->rq_xprt ) );
    nest_debug( NEST_DEBUG_NFS, "nfs req %s received on socket %d from %s.\n",
            req, rqstp->rq_xprt->xp_sock, ip ); 
    NestFree( ip );
}

int
filehandleToInt( NestFilehandle *handle ) {
    int i = (int)*handle;
    return i;
}

void *
filehandleCopy( void *nfs_fh, NestFilehandle *handle ) {
    memcpy( nfs_fh, handle, sizeof( NestFilehandle ) );
    return nfs_fh;
}

void *
filehandleCopy2( NestFilehandle *handle, void *nfs_fh ) {
    memcpy(handle, nfs_fh, sizeof( NestFilehandle));
    return handle; 
}

int
nfsFilehandleToInt( void *nfs ) {
    NestFilehandle nest;
    filehandleCopy2( &nest, nfs );
    return filehandleToInt( &nest );
}

enum nfsstat
NestStatusToNFSStatus( NestReplyStatus status ) {
    switch( status ) {
    case NEST_SUCCESS:                  return NFS_OK;

    case NEST_FILE_NOT_FOUND:
    case NEST_LOCAL_FILE_NOT_FOUND:     
    case NEST_REMOTE_FILE_NOT_FOUND:    return NFSERR_NOENT;

    case NEST_INSUFFICIENT_SPACE:
    case NEST_LOT_NOT_ENOUGH:           
    case NEST_DISK_FULL:                return NFSERR_NOSPC;

    case NEST_NOT_EMPTY:                return NFSERR_NOTEMPTY;

    case NEST_FILENAME_TOO_LONG:        return NFSERR_NAMETOOLONG;

    case NEST_FILE_EXISTS:              
    case NEST_ALREADY_EXISTS:           return NFSERR_EXIST;

                                        
    case NEST_STALE_FH:                 return NFSERR_STALE;

    case NEST_NONPRIVELEDGED_USER:      return NFSERR_PERM;

    case NEST_TEMPORARILY_UNAVAILABLE:  
    case NEST_NO_CONNECTION:            
    case NEST_BAD_PARAMETERS:           
    case NEST_NOT_YET_IMPLEMENTED:      
    case NEST_INTR:                     
    case NEST_UNKNOWN_REQUEST_TYPE:    
    default:
        nest_warn( __FILE__, __FUNCTION__, "Add a value for nest error %d:%s to %s:%s.\n",
                status, NestErrorString( status ), __FILE__, __FUNCTION__ );
        return NFSERR_IO;
    }
}

NestReplyStatus
NFSStatusToNestStatus( enum nfsstat status ) {
    switch( status ) {
    case NFS_OK:                return NEST_SUCCESS; 

    case NFSERR_PERM:           return NEST_NONPRIVELEDGED_USER; 
    case NFSERR_NOENT:          return NEST_FILE_NOT_FOUND; 
    case NFSERR_ACCES:          return NEST_NONPRIVELEDGED_USER;
    case NFSERR_EXIST:          return NEST_FILE_EXISTS; 
    case NFSERR_NAMETOOLONG:    return NEST_FILENAME_TOO_LONG; 
    case NFSERR_NOTEMPTY:       return NEST_NOT_EMPTY; 

    case NFSERR_NOSPC:          
    case NFSERR_DQUOT:          return NEST_DISK_FULL;  

    case NFSERR_STALE:          return NEST_STALE_FH;

    case NFSERR_IO:             
    case NFSERR_NXIO:          
    case NFSERR_NODEV:        
    case NFSERR_WFLUSH:        
    case NFSERR_NOTDIR:    
    case NFSERR_ISDIR:      
    case NFSERR_FBIG:        
    case NFSERR_ROFS:         
    default:
        nest_warn( __FILE__, __FUNCTION__, "Add a value for nfs error %d:%s to %s:%s.\n",
                (int)status, nfs_error( status ), __FILE__, __FUNCTION__ );
        return NEST_UNKNOWN_ERROR;
    }
}

char *
mount_request( int req ) {
    switch( req ) {
    case MOUNTPROC_NULL:        return "null";
    case MOUNTPROC_MNT:         return "mount";
    case MOUNTPROC_DUMP:        return "dump";
    case MOUNTPROC_UMNT:        return "unmount";
    case MOUNTPROC_UMNTALL:     return "unmountall";
    case MOUNTPROC_EXPORT:      return "export";
    case MOUNTPROC_EXPORTALL:   return "exportall";
    default:
        assert( 0 );
    }
}

char *
nfs_request( int req ) {
    switch( req ) {
    case NFSPROC_NULL:          return "null";
    case NFSPROC_GETATTR:       return "getattr";
    case NFSPROC_SETATTR:       return "setattr";
    case NFSPROC_ROOT:          return "root";
    case NFSPROC_LOOKUP:        return "lookup";
    case NFSPROC_READLINK:      return "readlink";
    case NFSPROC_READ:          return "read";
    case NFSPROC_WRITECACHE:    return "writecache";
    case NFSPROC_WRITE:         return "write";
    case NFSPROC_CREATE:        return "create";
    case NFSPROC_REMOVE:        return "remove";
    case NFSPROC_RENAME:        return "rename";
    case NFSPROC_LINK:          return "link";
    case NFSPROC_SYMLINK:       return "symlink";
    case NFSPROC_MKDIR:         return "mkdir";
    case NFSPROC_RMDIR:         return "rmdir";
    case NFSPROC_READDIR:       return "readdir";
    case NFSPROC_STATFS:        return "statfs";
    default:
        assert( 0 );
    }
} 


/*
 * NFS errors
 */
char *
nfs_error( enum nfsstat nfsstat ) {
    switch (nfsstat) {
    case NFS_OK:
       return "No error";
    case NFSERR_PERM:
       return "Not owner";
    case NFSERR_NOENT:
       return "No such file or directory";
    case NFSERR_IO:
       return "I/O error";
    case NFSERR_NXIO:
       return "No such device or address";
    case NFSERR_ACCES:
       return "Permission denied";
    case NFSERR_EXIST:
       return "File exists";
    case NFSERR_NODEV:
       return "No such device";
    case NFSERR_NOTDIR:
       return "Not a directory";
    case NFSERR_ISDIR:
       return "Is a directory";
    case NFSERR_FBIG:
       return "File too large";
    case NFSERR_NOSPC:
       return "No space left on device";
    case NFSERR_ROFS:
       return "Read-only file system";
    case NFSERR_NAMETOOLONG:
       return "File name too long";
    case NFSERR_NOTEMPTY:
       return "Directory not empty";
    case NFSERR_DQUOT:
       return "Disc quota exceeded";
    case NFSERR_STALE:
       return "Stale NFS file handle";
    case NFSERR_WFLUSH:
       return "Write cache flushed";
    default:
       return "UKNOWN NFS ERROR";
    }
}


